MODULE anti_bump

USE elements

implicit none

real, parameter, private :: SIGMADEF = 0.3

type bump_settings
  integer :: baction = 0     ! 0 = bump disabled, 1 = default strategy, 2 = user strategy
  real    :: weight  = 1     ! weight for bump in the CF
  real    :: scaledist = 0.7 ! scale factor on distance
end type bump_settings

integer, dimension(4), parameter, private :: abelem = [C_at,N_at,O_at,S_at]

!S set_antibump_auto(bump, btable, atom, frag)                     !Generate automatically list of anti-bumping restraints
!F function bump_distance(z1,z2) result(dist)                      !Set target distances as sum of vdW radius
!S create_bump_table(atom, bumptab)                                !Create table of anti-bump restraints from the atomic model
!S activate_bump(bumptab,el1,el2,targ)                             !Activate bump restraint in the bump table and set target distance 
!S bump_from_table(bumptab,bump,atom,frag,bset)                    !Set restraints list from bump table
!S table_from_bump(bumptab,bump,atom,scaled)                       !Set bump table from existing restraint list. Used to align directive and graphic

!S bump_from_string(string, atom, bump, frag, scaled, ier)         !Get anti-bump restraints from a string
!S set_antibump_atoms(bump,atom,cell,frag,vat1,vat2,scaled,targ)   !Set anti-bump restraints between atoms vat
!S set_antibump_spec(bump,atom,frag,spec,scaled,targ)              !Set anti-bump restraints for for species in the array spec
!S set_antibump_spec1(bump,atom,frag,spec1,spec2,scaled,targ)      !Set all bump restraints between species spec1 spec2
!S print_antibump(kpr,bump,atom)                                   !Print list of anti-bumping restraints
!F antibump_function(bump,atom,doc,kpr) result(bfunc)              !Compute cost function for anti-bump restraints

CONTAINS

   subroutine set_antibump_auto(bump, atom, cell, frag, scaled)
!
!  Generate automatically list of anti-bumping restraints
!
   USE atom_type_util
   USE fragmentmod
   USE unit_cell
   USE rrestr
   type(restraint_type), dimension(:), allocatable, intent(inout) :: bump
   type(atom_type), dimension(:), allocatable, intent(in)         :: atom
   type(cell_type), intent(in)                                    :: cell
   type(fragment_type), dimension(:), allocatable, intent(in)     :: frag
   real, intent(in)                                               :: scaled
!
   call set_antibump_spec(bump, atom, cell, frag, eleminfo(abelem)%lab, scaled)
!
   end subroutine set_antibump_auto

!--------------------------------------------------------------

   real function bump_distance(z1,z2) result(dist)
!
!  Set target distances as sum of vdW radius
!
   integer, intent(in) :: z1,z2
   real, parameter :: FACTH = 0.9
!
   if ((z1 == O_at .and. z2 == O_at) .or.      &   ! riduce distance for possible h-bond
       (z1 == N_at .and. z2 == O_at) .or.      &
       (z1 == O_at .and. z2 == N_at)) then     
       dist = FACTH*(eleminfo(z1)%w_radius + eleminfo(z2)%w_radius)
   else
       dist = eleminfo(z1)%w_radius + eleminfo(z2)%w_radius
   endif
!
   end function bump_distance

!--------------------------------------------------------------

   subroutine create_bump_table(atom, bumptab)
!
!  Create table of anti-bump restraints from the atomic model
!
   USE atom_type_util
   USE strutil
   USE arrayutil
   USE rrestr
   type(element_type), dimension(:), allocatable                :: elem
   type(atom_type), dimension(:), intent(in)                    :: atom
   type(restraint_type), dimension(:), allocatable, intent(out) :: bumptab
   integer                                                      :: nelem
   integer                                                      :: i,j
   integer                                                      :: nbump
!
   call elements_from_atom(atom,elem)
   nelem = numelem(elem)
   if (nelem > 0) then
       call resize_restraints(bumptab,(nelem*nelem + nelem)/2)
       nbump = 0
       do i=1,nelem
          do j=i,nelem
             nbump = nbump + 1
             call resize_array(bumptab(nbump)%na,2)
             bumptab(nbump)%code = -4  ! negative value for not-active anti-bump
             bumptab(nbump)%na(1) = elem(i)%ptab 
             bumptab(nbump)%na(2) = elem(j)%ptab 
             bumptab(nbump)%targ = bump_distance(elem(i)%z, elem(j)%z)
!             bumptab(nbump)%targ = SCALEDIST*(elem(i)%w_radius + elem(j)%w_radius)
!              write(0,*)trim(elem(i)%lab)//'-'//trim(elem(j)%lab),bumptab(nbump)%targ,bump_distance(elem(i)%ptab,elem(j)%ptab)
          enddo
       enddo
   endif
!
   end subroutine create_bump_table

!--------------------------------------------------------------

   subroutine activate_bump(bumptab,el1,el2,targ)
!
!  Activate bump restraint in the bump table and set target distance 
!
   USE rrestr
   type(restraint_type), dimension(:), intent(inout) :: bumptab  
   integer, intent(in)                               :: el1,el2 ! pointers to table of elements
   real, intent(in), optional                        :: targ    ! target distance
   integer                                           :: i
!
   do i=1,size(bumptab)
      if ((bumptab(i)%na(1) == el1 .and. bumptab(i)%na(2) == el2) .or.   &
          (bumptab(i)%na(1) == el2 .and. bumptab(i)%na(2) == el1)) then
           bumptab(i)%code = abs(bumptab(i)%code)
           if (present(targ)) bumptab(i)%targ = targ
      endif
   enddo
!
   end subroutine activate_bump
   
!--------------------------------------------------------------

   subroutine bump_from_table(bumptab,bump,atom,cell,frag,bset)
!
!  Set restraints list from bump table
!
   USE fragmentmod
   USE unit_cell
   USE rrestr
   type(restraint_type), dimension(:), intent(in)                 :: bumptab
   type(restraint_type), dimension(:), allocatable, intent(inout) :: bump
   type(atom_type), dimension(:), allocatable, intent(in)         :: atom
   type(cell_type), intent(in)                                    :: cell
   type(fragment_type), dimension(:), allocatable, intent(in)     :: frag
   type(bump_settings)                                            :: bset
   integer                                                        :: i
!
   do i=1,size(bumptab)
      if (bumptab(i)%code > 0) then
          call set_antibump_spec1(bump, atom, cell, frag,   &
          eleminfo(bumptab(i)%na(1))%lab, eleminfo(bumptab(i)%na(2))%lab, bset%scaledist, bumptab(i)%targ)
      endif
   enddo
!
   end subroutine bump_from_table

!--------------------------------------------------------------

   subroutine table_from_bump(bumptab,bump,atom,bset)
!
!  Set bump table from existing restraint list. Used to align directive and graphic
!
   USE atom_type_util
   USE rrestr
   type(restraint_type), dimension(:), allocatable, intent(inout) :: bumptab
   type(restraint_type), dimension(:), allocatable, intent(in)    :: bump
   type(atom_type), dimension(:), allocatable, intent(in)         :: atom
   type(bump_settings), intent(inout)                             :: bset
   integer                                                        :: i,j,i1,i2,pos
   integer                                                        :: nat1,nat2
   integer, dimension(size(atom))                                 :: vat1,vat2
   logical                                                        :: ltarg
   real                                                           :: targ
!
   if (nrestraints(bump) == 0) return
!
   loop_bump: do i=1,size(bumptab)
!
!     Get atoms of specie na(1) and na(2)
      call get_atoms_of_specie(bumptab(i)%na(1),atom,vat1,nat1)
      if (nat1 == 0) cycle
      call get_atoms_of_specie(bumptab(i)%na(2),atom,vat2,nat2)
      if (nat2 == 0) cycle
!
!     Check if all atoms are involved in restraints
      ltarg = .false.
      do i1=1,nat1
         do i2=1,nat2
            pos = restraint_position(bump,[abs(vat1(i1)),vat2(i2)])
            if (pos == 0) cycle loop_bump
!corr            if (pos == 0) then
!corr                pos = restraint_position(bump,restraint_type(code=ABUMP,na=[-vat1(i1),vat2(i2)]))
!corr                if (pos == 0) cycle loop_bump
!corr            endif
!
!           Check target distance, should be the same
            if (.not.ltarg) then
                targ = bump(pos)%targ
                ltarg = .true.
            else
                if (targ /= bump(pos)%targ) cycle loop_bump
            endif
         enddo
      enddo
      call activate_bump(bumptab,bumptab(i)%na(1),bumptab(i)%na(2),targ/bset%scaledist)
   enddo loop_bump
!
!  Set type of action
   bset%baction = 2 
   loop_elem: do i=1,size(abelem)
      do j=i,size(abelem)
         if ((numatomspec(atom,abelem(i)) > 0) .and. (numatomspec(atom,abelem(j))) > 0) then ! elements are present in atom?
             pos = restraint_position(bumptab,[abelem(i),abelem(j)])
             if (pos == 0) exit loop_elem  ! this not generally possible
             if (bumptab(pos)%code < 0) then
                 bset%baction = 1 
                 exit loop_elem
             endif
         endif
      enddo
   enddo loop_elem
!
   end subroutine table_from_bump

!--------------------------------------------------------------

   subroutine bump_from_string(string, atom, cell, bump, frag, scaled, ier)
!
!  Get anti-bump restraints from a string
!
   USE strutil
   !USE atom_basic, only:atom_type
   USE atom_type_util
   USE fragmentmod
   USE unit_cell
   USE rrestr
   character(len=*), intent(in)                                   :: string
   type(atom_type), dimension(:), allocatable, intent(in)         :: atom
   type(cell_type), intent(in)                                    :: cell
   type(restraint_type), dimension(:), allocatable, intent(inout) :: bump
   type(fragment_type), dimension(:), allocatable, intent(in)     :: frag
   real, intent(in)                                               :: scaled
   integer, intent(out)                                           :: ier
   character(len=len_trim(string))                                :: line
   integer                                                        :: nspec
   character(len=20), dimension(4)                                :: wordv
   integer                                                        :: nword
   logical                                                        :: ok_bump 
   integer                                                        :: i
   real                                                           :: val, targ
   integer                                                        :: lenv
   integer                                                        :: irep
   integer, dimension(10)                                         :: ivet
   real, dimension(10)                                            :: vet
   integer                                                        :: iv
   integer                                                        :: nfrag
   integer, dimension(size(atom)) :: vat1,vat2
   integer :: nat1,nat2
!
   ier = 0
   ok_bump = .true.
   line = trim(adjustl(string))
   call s_filter(line)
   if (len_trim(line) /= 0) then
       call s_rep(line,'frag',' ',irep)
       select case(irep) 
         case (0)    ! set bump as: bump specie1 specie2 target_distance
           call get_words(line,wordv,nword)
           if (nword >=2 ) then
               nspec = 0
               targ = -2
               do i=1,nword
                  if (nspec == 2) then
                      ier = s_to_r(wordv(i),val,lenv)
                      if (ier == 0) targ = val
                      exit
                  else
                      nspec = nspec + 1
                  endif
               enddo
               if (nspec == 2) then
                   call get_atoms_of_string(wordv(1),atom,vat1,nat1)
                   if (nat1 > 0) then
                       call get_atoms_of_string(wordv(2),atom,vat2,nat2)
                       if (nat2 > 0) then
                           call set_antibump_atoms(bump,atom,cell,frag,vat1(:nat1),vat2(:nat2),scaled,targ)
                       endif
                   endif
                   if (nat1 == 0 .or. nat2 == 0) ier = 1
               endif
           endif

         case (1)  
           ier = 2

         case (2)   ! es. frag1 frag2 3.0
           nfrag = numfragments(frag)
           call getnum(line,vet,ivet,iv)
           select case(iv)       
             case (2,3)
               if (ivet(1) > nfrag .or. ivet(2) > nfrag .or. ivet(1) < 1 .or. ivet(2) < 1) then
                   ier = 3
               else
                   if (iv == 2) targ = -2
                   call set_antibump_atoms(bump,atom,cell,frag,frag(ivet(1))%pos,frag(ivet(2))%pos,scaled,targ)
               endif
             case default
               ier = 3
           end select

      end select
   else
      call set_antibump_auto(bump, atom, cell, frag, scaled)
   endif
!
   end subroutine bump_from_string

!--------------------------------------------------------------

   subroutine set_antibump_atoms(bump,atom,cell,frag,vat1,vat2,scaled,targ)
!
!  Set anti-bump restraints between atoms vat
!
   USE atom_type_util
   USE fragmentmod
   USE cgeom
   USE unit_cell
   USE rrestr
   type(restraint_type), dimension(:), allocatable, intent(inout) :: bump
   type(atom_type), dimension(:), allocatable, intent(in)         :: atom
   type(cell_type), intent(in)                                    :: cell
   type(fragment_type), dimension(:), allocatable, intent(in)     :: frag
   integer, dimension(:), intent(in)                              :: vat1,vat2
   !character(len=*), dimension(:), intent(in)                     :: spec
   real, intent(in)                                               :: scaled
   real, intent(in), optional                                     :: targ
   integer                                                        :: nbump, nat
   integer                                                        :: i,j
   type(restraint_type)                                           :: bump_new
   integer                                                        :: respos
   integer :: a1,a2,nbumpold
   logical :: ltarg
!
   nbump = nrestraints(bump)
   if (size(vat1)*size(vat2) == 0) return
!
   ltarg = .false.
   if (present(targ)) then
       ltarg =  targ > 0
   endif
!
!  allocate bump to a reasonable dimension
   !call resize_restraints(bump,nbump + size(vat)*(size(vat) -1)/2)   ! n*(n-1)/2 is sum of serie an=1,2,3,..,n
   call resize_restraints(bump,nbump + size(vat1)*size(vat2))
!   
   nat = numatoms(atom)
   nbumpold = nbump
   bump_new%code = ABUMP
   allocate(bump_new%na(2))
!
   do i=1, size(vat1)
      do j=1, size(vat2)
         a1 = vat1(i)
         a2 = vat2(j)
!
!        Check existence of new restraint
         !bump_new = restraint_type(ABUMP,[a1,a2])
         bump_new%na = [a1,a2]
         if (fragment_pos(frag,a1) == fragment_pos(frag,a2)) then                  ! if in the same fragment
             bump_new%na(1) = -a1 ! only equivalent position are checked for atoms in the same molecular fragment
         endif
         respos = restraint_position(bump,bump_new)
         if (respos > nbumpold) cycle ! a1-a2 duplicated in  arrays vat1-vat2

         bump_new%val = distanzaC(atom(a1)%xc,atom(a2)%xc,cell%get_g())
         if (ltarg) then
             bump_new%targ = targ * scaled
         else
             bump_new%targ = bump_distance(atom(a1)%z(), atom(a2)%z()) * scaled
         endif
         bump_new%sigma = SIGMADEF

         if (respos == 0) then  ! add anti-bump in a list
             nbump = nbump + 1
             bump(nbump) = bump_new
         else                   ! replace existing anti-bump
             bump(respos) = bump_new
         endif
      enddo
   enddo
!
   call resize_restraints(bump,nbump)
!
   end subroutine set_antibump_atoms

!--------------------------------------------------------------

   subroutine set_antibump_spec(bump,atom,cell,frag,spec,scaled,targ)
!
!  Set anti-bump restraints for species in the array spec
!
   USE atom_type_util
   USE fragmentmod
   USE unit_cell
   USE rrestr
   type(restraint_type), dimension(:), allocatable, intent(inout) :: bump
   type(atom_type), dimension(:), allocatable, intent(in)         :: atom
   type(cell_type), intent(in)                                    :: cell
   type(fragment_type), dimension(:), allocatable, intent(in)     :: frag
   character(len=*), dimension(:), intent(in)                     :: spec
   real, intent(in)                                               :: scaled
   real, intent(in), optional                                     :: targ
   integer                                                        :: i,j
   integer, dimension(size(atom)) :: vat1,vat2
   integer                        :: natom1,natom2
!
   do i=1,size(spec)
      call get_atoms_of_specie(pxen_from_specie(spec(i)),atom,vat1,natom1)
      if (natom1 == 0) cycle
      call set_antibump_atoms(bump,atom,cell,frag,vat1(:natom1),vat1(:natom1),scaled,targ)
      do j=i+1,size(spec)
         call get_atoms_of_specie(pxen_from_specie(spec(j)),atom,vat2,natom2)
         if (natom2 == 0) cycle
         call set_antibump_atoms(bump,atom,cell,frag,vat1(:natom1),vat2(:natom2),scaled,targ)
      enddo
   enddo
!
   end subroutine set_antibump_spec

!--------------------------------------------------------------

   subroutine set_antibump_spec1(bump,atom,cell,frag,spec1,spec2,scaled,targ)
!
!  Set all bump restraints between species spec1 spec2
!
   USE atom_type_util
   USE fragmentmod
   USE unit_cell
   USE rrestr
   type(restraint_type), dimension(:), allocatable, intent(inout) :: bump
   type(atom_type), dimension(:), allocatable, intent(in)         :: atom
   type(cell_type), intent(in)                                    :: cell
   type(fragment_type), dimension(:), allocatable, intent(in)     :: frag
   character(len=*), intent(in)                                   :: spec1, spec2
   real, intent(in)                                               :: scaled
   real, intent(in), optional                                     :: targ
   integer, dimension(size(atom))                                 :: vat1,vat2
   integer                                                        :: natom1,natom2
!
   call get_atoms_of_specie(pxen_from_specie(spec1),atom,vat1,natom1)
   if (natom1 == 0) return
   call get_atoms_of_specie(pxen_from_specie(spec2),atom,vat2,natom2)
   if (natom2 == 0) return
   call set_antibump_atoms(bump,atom,cell,frag,vat1(:natom1),vat2(:natom2),scaled,targ)
!
   end subroutine set_antibump_spec1
   
!--------------------------------------------------------------

   subroutine print_antibump(kpr,bump,atom,cell)
!
!  Print list of anti-bumping restraints
!
   USE atom_type_util
   USE connect_mod
   USE unit_cell
   USE rrestr
   integer, intent(in)                                    :: kpr
   type(restraint_type), dimension(:), allocatable, intent(in) :: bump
   type(atom_type), dimension(:), allocatable, intent(in) :: atom
   type(cell_type), intent(in)                            :: cell
   if (nrestraints(bump) > 0) then
       call print_restraints(kpr,bump,ABUMP,atom,cell)
   endif
   end subroutine print_antibump

!--------------------------------------------------------------

   real function antibump_function(bump,atom,spg,gmat,kpr) result(bfunc)
!
!  Compute cost function for anti-bump restraints and update current value bump%val
!
   USE atom_type_util
   USE fragmentmod
   USE connect_mod
   USE cgeom
   USE rrestr
   USE spginfom
   type(restraint_type), dimension(:), allocatable, intent(inout) :: bump
   type(atom_type), dimension(:), allocatable, intent(in)         :: atom
   type(spaceg_type), intent(in)                                  :: spg
   real, dimension(3,3), intent(in)                               :: gmat
   logical, intent(in)                                            :: kpr
   integer                                                        :: nbump
   integer                                                        :: n1, n2
   real                                                           :: dd
   type(atom_type)                                                :: at
   integer                                                        :: i, k
   real, dimension(3)                                             :: ktra
   logical                                                        :: same_frag
   real                                                           :: distmin
   real, parameter                                                :: EPS = epsilon(1.0)
   real, parameter                                                :: DDOC = 1.0
!
   bfunc = 0
   nbump = nrestraints(bump)
   do i=1,nbump
      n1 = bump(i)%na(1)
      n2 = bump(i)%na(2)
      if (n1 < 0) then
          n1 = abs(n1)
          same_frag = .true.
      else
          same_frag = .false.
      endif
      distmin = huge(1.0) !!!!distanzaC(atom(n2)%xc,atom(n1)%xc,gmat)
      do k=1,spg%nsymop
         at = atom_symm(atom(n1),spg%symop(k))
         call xdisteqs(atom(n2)%xc,at%xc,gmat,dd,ktra)
       !write(0,*)'BUMP=',i,k,distmin,dd
         if (same_frag .and. k == 1) then           ! n1 and n2 are in the same fragment when k==1
!
!            check if cell translation is 0: ktra - (at%xc-atom(n2)%xc) == 0
             if (atom(n2) + ktra == at) cycle  ! if true n1 is not an equivalent for translation
         endif
         if (dd < distmin) distmin = dd
         if (dd < bump(i)%targ) then
!
!            Check for doc: same elements, doc active, dist < DISTMIN
             if ((atom(n1)%doc) .and.  (atom(n2)%doc) .and.          &
                 (atom(n1)%ptab == atom(n2)%ptab) .and. (dd < DDOC)) then             
                  !write(0,*)'ocry=',trim(atom(n1)%lab)//'-'//trim(atom(n2)%lab),dd
                  cycle
             endif
!
             bfunc = bfunc + ((1/bump(i)%sigma)*(dd - bump(i)%targ))**4 
             if(kpr) then
                write(71,*)'activate bump:',atom(n1)%lab,atom(n2)%lab,dd,(1/bump(i)%sigma)*(dd - bump(i)%targ)**2
             endif
         endif
      enddo
      bump(i)%val = distmin
   enddo
   if (kpr) then
       write(71,*)'fbump=',bfunc
   endif
!
   end function antibump_function

END MODULE anti_bump

!corr   subroutine test_bump()
!corr   USE variables
!corr   USE fragmentmod
!corr   USE connect_mod
!corr   USE anti_bump
!corr   USE General, only: cell
!corr   USE unit_cell
!corr   type(fragment_type), dimension(:), allocatable :: frag
!corr   type(restraint_type), dimension(:), allocatable     :: bump
!corr!corr   type(bump_settings) :: bset
!corr   integer :: nfrag
!corr   real :: fbump
!corr   type(cell_type) :: cellt
!corr          write(0,*)'start test_bump',size(atm),size(lconn)
!corr   cellt = set_cell_type(cell)
!corr   call get_fragments(atm,cellt,lconn,nfrag,frag)
!corr   !call set_antibump_auto(bump, atm, frag)
!corr        !call set_antibump_spec1(bump, atm, cellt, frag, 'C', 'C',bset%scaledist)
!corr        !call set_antibump_spec1(bump, atm, cellt, frag, 'C', 'H',bset%scaledist)
!corr        !call set_antibump_spec1(bump, atm, cellt, frag, 'O', 'O',bset%scaledist)
!corr        !call set_antibump_spec1(bump, atm, frag, 'C', 'O', 2.4)
!corr        !call set_antibump_spec1(bump, atm, frag, 'O', 'O', 2.3)
!corr        !call set_antibump_spec1(bump, atm, frag, 'V', 'C')
!corr        !call set_antibump_spec1(bump, atm, frag, 'V', 'C')
!corr        !call set_antibump_spec1(bump, atm, frag, 'O', 'O',2.0)
!corr   !call set_antibump_spec1(bump, atm, frag, 'C', 'H')
!corr   !call set_antibump_spec1(bump, atm, frag, 'H', 'H')
!corr   !call set_antibump_spec1(bump, atm, frag, 'V', 'C')
!corr   !call set_antibump_spec1(bump, atm, frag, 'V', 'V')
!corr   !call set_antibump_spec1(bump, atm, frag, 'O', 'H')  !-- prosssimo test
!corr   fbump = antibump_function(bump, atm, cellt%get_g(),.false.,.false.)
!corr   call print_antibump(0,bump,atm,cellt)
!corr   fbump = antibump_function(bump, atm, cellt%get_g(),.false., .true.)
!corr   end subroutine test_bump
