MODULE datasetmod

   USE pointmod, only: point_type
   USE background, only: back_condition_type

   implicit none

   integer, parameter :: POW_DATA=1,SCRYS_DATA=2

   type dataset_type
     real, dimension(:), allocatable :: x        ! observed x
     real, dimension(:), allocatable :: x0       ! observed corrected
     real, dimension(:), allocatable :: y        ! observed y
     real, dimension(:), allocatable :: yc       ! calculated on all crystal phases
     real, dimension(3)              :: wave
     integer                         :: nwave = 0
     real                            :: ratio = 0.5
     integer                         :: radtype
     logical                         :: sync = .false.
     character(len=256)              :: fname = ' '
     real                            :: tstep,tmin,tmax
     integer                         :: datatype = POW_DATA
     integer                         :: nc1,nc2  ! range for calculated (yb and yc)

     real, dimension(:), allocatable             :: yb    ! background
     type(back_condition_type)                   :: cond    ! settings for create backgrund
     type(point_type), dimension(:), allocatable :: points
     real                                        :: thzerob=0.0 ! zero of polynomial
     real, dimension(:), allocatable             :: coef    ! background coefficients

   contains

     ! Data retrieval methods
     procedure          :: open_file
     procedure          :: data_read => read_data_bin_s
     procedure          :: npoints => get_npoints
     procedure          :: npoints_back
     procedure, private :: get_points_p
     procedure, private :: get_points_a
     generic            :: get_points => get_points_a, get_points_p
     procedure          :: get_coef
     procedure          :: zeroback
     procedure          :: get_cond
     procedure          :: set_cond
     procedure          :: set_limit

     ! Data modification methods
     procedure          :: set_data
     procedure          :: resize
     procedure          :: set_points
     procedure          :: make_background

     ! Output
     procedure          :: data_save => save_data_bin_s
     procedure          :: prn => print_info

   end type dataset_type

CONTAINS

   integer function get_npoints(datas)
   USE arrayutil
   class(dataset_type), intent(in) :: datas
   get_npoints = size_array(datas%x)
   end function get_npoints 

!-------------------------------------------------------------------------

   subroutine open_file(this,filename,ier,wavel,filetyp,sync)
!
!  Load dataset from external file
!
   USE prog_constants
   USE filereading
   USE fileutil
   USE counts
   USE elements, only: RX_SOURCE,NEUTRON_SOURCE
   class(dataset_type), intent(inout)          :: this
   character(len=*), intent(in)                :: filename
   integer, intent(out)                        :: ier
   real, intent(in), optional                  :: wavel   ! wavel <= 0 will be ignored
   integer, intent(in), optional               :: filetyp
   logical, intent(in), optional               :: sync
   integer                                     :: count_type
   integer                                     :: ftype
   real                                        :: thmino, thstepo, thmaxo, wavelen
   integer                                     :: ncount
   type(point_type), allocatable, dimension(:) :: countp
   type(error_type)                            :: error
   integer                                     :: irad
!
   if (present(filetyp)) then
       ftype = filetyp
   else
       ftype = filetype_from_filename(filename)
   endif
   call read_datafile(filename,ftype,thmino,    &
        thstepo,thmaxo,countp,ncount,.false.,wavelen,count_type,error)
   if (error%signal) then
       call error%print()
       ier = 1
   else
       ier = 0
       this%radtype = RX_SOURCE
       if (present(sync)) then
           this%sync = sync
       else
           this%sync = .false.
       endif
       this%tstep = thstepo
       this%tmin = thmino
       this%tmax = thmaxo
       this%nwave = 1
       if (present(wavel)) then
           if (wavel > 0) wavelen = wavel
       endif
       if (wavelen <= 0) then
           this%wave(1) = DEF_WAVE
           call openwave(this%wave(1),irad)
           select case (irad)
             case (2)
               this%sync = .true.
             case (3)
               this%radtype = NEUTRON_SOURCE
           end select
       else
           this%wave(1) = wavelen
       endif
       if (count_type == COUNT_D_TYPE) countp = thvalue(countp,this%wave(1))
       call this%set_data(countp(:)%x,countp(:)%y,size(countp),filename,POW_DATA)
   endif
   end subroutine open_file

!-------------------------------------------------------------------------

   subroutine set_data(this,x,y,np,fname,datatype,x0)
   USE arrayutil
   class(dataset_type), intent(inout)       :: this
   real, dimension(:), intent(in)           :: x,y
   integer, intent(in)                      :: np
   character(len=*), intent(in)             :: fname
   integer, intent(in)                      :: datatype
   real, dimension(:), optional, intent(in) :: x0
!
   this%datatype = datatype
   this%fname = fname
   if (np > 0) then
       call new_array(this%x,np)
       call new_array(this%x0,np)
       call new_array(this%y,np)
       this%x(:) = x(:np)
       this%y(:) = y(:np)
       this%tmin = this%x(1)
       this%tmax = this%x(np)
       this%nc1 = 1
       this%nc2 = size(x)
       if (present(x0)) then
           this%x0 = x0
       else
           this%x0 = x
       endif
   endif
!
   end subroutine set_data

!-------------------------------------------------------------------------

   subroutine save_data_bin_s(this,unitbin)
   USE arrayutil
   class(dataset_type), intent(in) :: this
   integer, intent(in)             :: unitbin
   integer                         :: np,npyc,npb,npbp,npc
!
   np = this%npoints()
   write(unitbin) np
   if (np > 0) then
       npyc = size_array(this%yc)
       npb = size_array(this%yb)
       npbp = this%npoints_back()
       npc = size_array(this%coef)
       write(unitbin) this%x,this%x0,this%y,this%wave,this%nwave,this%ratio,this%tstep,    &
                      this%tmin,this%tmax,this%radtype,this%sync,this%fname,this%datatype, &
                      this%nc1,this%nc2,this%cond,this%thzerob,npyc,npb,npbp,npc
       if (npyc > 0) then
           write(unitbin)this%yc
       endif
       if (npb > 0) then
           write(unitbin)this%yb
       endif
       if (npbp > 0) then
           write(unitbin)this%points
       endif
       if (npc > 0) then
           write(unitbin)this%coef
       endif
   endif
!
   end subroutine save_data_bin_s

!-------------------------------------------------------------------------
   
   subroutine save_data_bin(datas,unitbin)
   type(dataset_type), dimension(:), allocatable, intent(in) :: datas
   integer, intent(in)                                       :: unitbin
   integer :: i
!
   write(unitbin) ndataset(datas)
   do i=1,ndataset(datas)
      call datas(i)%data_save(unitbin)
   enddo
!
   end subroutine save_data_bin 

!-------------------------------------------------------------------------

   subroutine read_data_bin_s(this, unitbin, err)
   USE arrayutil
   USE errormod
   USE pointmod
   class(dataset_type), intent(inout) :: this
   integer, intent(in)                :: unitbin
   type(error_type), intent(out)      :: err
   integer                            :: ier
   integer                            :: np,npyc,npb,npbp,npc
!
   read(unitbin,iostat=ier,err=10) np
   if (np > 0) then
       call new_array(this%x,np)
       call new_array(this%x0,np)
       call new_array(this%y,np)
       read(unitbin,iostat=ier,err=10)                                                     &
                      this%x,this%x0,this%y,this%wave,this%nwave,this%ratio,this%tstep,    &
                      this%tmin,this%tmax,this%radtype,this%sync,this%fname,this%datatype, &
                      this%nc1,this%nc2,this%cond,this%thzerob,npyc,npb,npbp,npc
       if (npyc > 0) then
           call new_array(this%yc,npyc)
           read(unitbin,iostat=ier,err=10)this%yc
       endif
       if (npb > 0) then
           call new_array(this%yb,npb)
           read(unitbin,iostat=ier,err=10)this%yb
       endif
       if (npbp > 0) then
           call new_points(this%points,npbp)
           read(unitbin,iostat=ier,err=10)this%points
       endif
       if (npc > 0) then
           call new_array(this%coef,npc)
           read(unitbin,iostat=ier,err=10)this%coef
       endif
   endif

10 continue
   if (ier /= 0) then
       call err%set('Error on reading diffraction data')
   endif
   end subroutine read_data_bin_s

!-------------------------------------------------------------------------

   subroutine read_data_bin(datas,unitbin,err)
   USE errormod
   type(dataset_type), intent(inout), dimension(:), allocatable :: datas
   integer, intent(in)                                          :: unitbin
   type(error_type), intent(out)                                :: err
   integer                                                      :: i,ier,ndat

   read(unitbin,iostat=ier) ndat
   if (ier /= 0) then
       call err%set("Error on reading datasets")
       return
   endif
! 
   call new_dataset(datas,ndat)
   do i=1,ndat
      call datas(i)%data_read(unitbin,err)
      if (err%signal) return
   enddo

   end subroutine read_data_bin

!-------------------------------------------------------------------------

   subroutine resize(this,xmin,xmax)
   USE arrayutil
   class(dataset_type), intent(inout) :: this
   real, intent(in), optional         :: xmin,xmax
   integer                            :: ipmin,ipmax,olddim,newdim
!
   olddim = size_array(this%x)
   if (olddim == 0) return

   if (present(xmin)) then
       ipmin = clocate(this%x,xmin)                  ! locazione del punto + vicino
       if (this%x(ipmin) < xmin) ipmin = ipmin + 1   ! ma non deve essere < di xmin
   else
       ipmin = 1
   endif
   if (present(xmax)) then
       ipmax = clocate(this%x,xmax)                  ! locazione del punto + vicino
       if (this%x(ipmax) > xmax) ipmax = ipmax - 1   ! ma non deve essere > di xmax
   else
       ipmax = olddim
   endif
   newdim = ipmax - ipmin + 1
   if (newdim > 0 .and. newdim /= olddim) then
       this%x(1:newdim) = this%x(ipmin:ipmax)
       this%x0(1:newdim) = this%x0(ipmin:ipmax)
       this%y(1:newdim) = this%y(ipmin:ipmax)
       call resize_array(this%x,newdim)
       call resize_array(this%x0,newdim)
       call resize_array(this%y,newdim)
       call this%set_limit(n2=newdim)
!corr       this%nc1 = 1
!corr       this%nc2 = size(x)
!corr!
!corr       if (allocated(this%yb)) then
!corr           this%yb(1:newdim) = this%yb(ipmin:ipmax)
!corr           call resize_array(this%yb,newdim)
!corr       endif
!corr       if (allocated(this%yc)) then
!corr           this%yc(1:newdim) = this%yc(ipmin:ipmax)
!corr           call resize_array(this%yc,newdim)
!corr       endif
       this%tmin = this%x(1)
       this%tmax = this%x(newdim)
   endif
!
   end subroutine resize

!----------------------------------------------------------------------------------------------

   subroutine print_info(this,kpr)
   USE arrayutil
   class(dataset_type), intent(in) :: this
   integer, intent(in)             :: kpr
   write(kpr,*)'Ncounts: ',size_array(this%x)
   end subroutine print_info

!----------------------------------------------------------------------------------------------

   integer function ndataset(datas)
   type(dataset_type), allocatable, intent(in) :: datas(:)
!
   if (allocated(datas)) then
       ndataset = size(datas)
   else
       ndataset = 0
   endif
!
   end function ndataset

!----------------------------------------------------------------------------------------------

   subroutine resize_dataset(datas,n)
!
!  Rialloca ad n un vettore reale.
!  Se savevet = .true. o non esiste si salva il suo contenuto.
!
   type(dataset_type), allocatable, intent(inout) :: datas(:)
   integer, intent(in)                            :: n
   integer                                        :: nv
   type(dataset_type), allocatable                :: vsav(:)
   integer                                        :: nsav
!
!  se n = 0 (riallocazione a 0): dealloca ed esci
   if (n == 0) then
       if (allocated(datas)) deallocate(datas)
       return
   endif
!
   if (.not.allocated(datas)) then
       allocate(datas(n))
   else
!
       nv = size(datas)
!
!      nsav contiene qual è la porzione di datas da salvare
       select case(nv-n)
         case (1:)       ! compatta x ad n
           nsav = n
         case (:-1)      ! espandi x ad n
           nsav = nv
         case (0)
           return        ! n=nv non fare niente
       end select
       allocate(vsav(n))
       vsav(:nsav) = datas(:nsav)
       call move_alloc(vsav,datas)
   endif
!
   end subroutine resize_dataset

!----------------------------------------------------------------------------------------------------

   subroutine new_dataset(datas,n)
!
!  Create new atoms
!
   type(dataset_type), allocatable, intent(inout) :: datas(:)
   integer, intent(in)                            :: n

   if (n < 0) return
   if (ndataset(datas) /= n) then
       if (allocated(datas))deallocate(datas)
       if (n > 0) allocate(datas(n))
   endif

   end subroutine new_dataset

!----------------------------------------------------------------------------------------------------

   subroutine clear_dataset(datas)
!
!  Delete all phases
!
   type(dataset_type), allocatable, intent(inout) :: datas(:)

   if (allocated(datas)) deallocate(datas)

   end subroutine clear_dataset

!----------------------------------------------------------------------------------------------------

   subroutine push_back_dataset(datas,val)
!
!  Adds a new phase at the end of the array
!
   type(dataset_type), allocatable, intent(inout) :: datas(:)
   type(dataset_type), intent(in)                 :: val
   integer                                        :: ndim
   ndim = ndataset(datas)
   call resize_dataset(datas,ndim+1)
   datas(ndim+1) = val
   end subroutine push_back_dataset

!----------------------------------------------------------------------------------------

   integer function npoints_back(datas)
   class(dataset_type), intent(in)                         :: datas
   if (allocated(datas%points)) then
       npoints_back = size(datas%points)
   else
       npoints_back = 0
   endif
   end function npoints_back

!----------------------------------------------------------------------------------------

   subroutine get_points_p(datas,points)
!
!  Get background points
!
   USE pointmod
   class(dataset_type), intent(in)                          :: datas
   type(point_type), dimension(:), allocatable, intent(out) :: points
   call copy_points(points,datas%points)
   end subroutine get_points_p

!----------------------------------------------------------------------------------------

   subroutine get_points_a(datas,vet)
!
!  Get background points
!
   USE pointmod
   USE arrayutil
   class(dataset_type), intent(in)                :: datas
   real, dimension(:,:), allocatable, intent(out) :: vet
   integer                                        :: np,i
   np = datas%npoints_back()
   if (np > 0) then
       call new_array(vet,[1,1],[np,2])
       do i=1,np
          vet(i,1) = datas%points(i)%x
          vet(i,2) = datas%points(i)%y
       enddo
   endif
   end subroutine get_points_a

!----------------------------------------------------------------------------------------

   subroutine set_points(datas,points)
!
!  Get background points
!
   USE pointmod
   class(dataset_type), intent(inout)                      :: datas
   type(point_type), dimension(:), allocatable, intent(in) :: points
   call copy_points(datas%points,points)
   end subroutine set_points

!----------------------------------------------------------------------------------------

   subroutine make_background(datas)  !!!,nps,npe)
   USE arrayutil
   USE background, only: NMAXCOEFB, genera_background
   class(dataset_type), intent(inout)           :: datas
!corr   integer, intent(in), optional                :: nps,npe
   integer                                      :: npb
!corr   integer :: npstart,npend
   call resize_array(datas%coef,NMAXCOEFB)
!corr   if (present(nps)) then
!corr       npstart = nps
!corr   else
!corr       npstart = 1
!corr   endif
!corr   if (present(npe)) then
!corr       npend = npe
!corr   else
!corr       npend = size(datas%y)
!corr   endif
   call new_array(datas%yb,datas%nc2-datas%nc1+1)
   call genera_background(datas%x0(datas%nc1:datas%nc2),datas%y(datas%nc1:datas%nc2),datas%yb,   &
                          datas%coef,datas%thzerob,datas%points,npb,datas%cond)
   end subroutine make_background

!----------------------------------------------------------------------------------------

   subroutine get_coef(datas,coef)  !!!!FIXME - add allocation on coef
   class(dataset_type), intent(in) :: datas
   real, dimension(:), intent(out)    :: coef
   coef = datas%coef
   end subroutine get_coef

!----------------------------------------------------------------------------------------

   real function zeroback(datas)
   class(dataset_type), intent(in) :: datas
   zeroback = datas%thzerob
   end function zeroback

!----------------------------------------------------------------------------------------

   subroutine set_limit(datas,n1,n2)
!
!  Set limit for calculate. Default is lbound(x):ubound(x)
!
   USE arrayutil
   class(dataset_type), intent(inout) :: datas
   integer, intent(in), optional      :: n1,n2
   integer                            :: nc1,nc2
!
   if (present(n1)) then
       nc1 = n1
   else
       nc1 = lbound(datas%x,dim=1)
   endif
   if (present(n2)) then
       nc2 = n2
   else
       nc2 = ubound(datas%x,dim=1)
   endif
   datas%nc1 = nc1
   datas%nc2 = nc2
!
   if (allocated(datas%yb)) call resize_array(datas%yb,nc2-nc1+1)
   if (allocated(datas%yc)) call resize_array(datas%yc,nc2-nc1+1)
!
   end subroutine set_limit

!----------------------------------------------------------------------------------------

   subroutine set_cond(datas,cond)
!
!  Set conditions for background
!
   class(dataset_type), intent(inout) :: datas
   type(back_condition_type), intent(in) :: cond
   datas%cond = cond
   end subroutine set_cond

!----------------------------------------------------------------------------------------

   subroutine get_cond(datas,cond)
!
!  Get conditions for background
!
   class(dataset_type), intent(in)     :: datas
   type(back_condition_type), intent(out) :: cond
   cond = datas%cond
   end subroutine get_cond

END MODULE datasetmod
