module psvoigtf
  USE progtype, only: rparam_type
  USE trig_constants, only: pi,dtor,rtod

  implicit none 

  type pvoigt_function_type
     type(rparam_type) :: u,v,w
     type(rparam_type) :: eta0,eta1,eta2
     type(rparam_type) :: asym
  end type pvoigt_function_type

  real, parameter, private :: GAUSCONST2 = 4.0*alog(2.0)
  real, parameter, private :: GAUSCONST = sqrt(GAUSCONST2/pi)
  real, parameter, private :: CONSTLOREN=2.0/pi

  real, dimension(:), allocatable, private :: etav
  real, dimension(:), allocatable, private :: ruvw

contains 

   subroutine init_calc(n)
   USE arrayutil
   integer, intent(in)          :: n
!
   if (allocated(etav) .and. n == size(etav)) return
   call new_array(etav,n)
   call new_array(ruvw,n)
!
   end subroutine init_calc

!----------------------------------------------------------------------------------------------------

   function init_psvoigt_param()  result(param)
   USE progtype, only: rparam_type
   type(pvoigt_function_type) :: param

   param%u = rparam_type(val=0.0,str='U')
   param%v = rparam_type(val=0.0,str='V')
   param%w = rparam_type(val=0.01,str='W')

   param%eta0 = rparam_type(val=0.5,str='eta0')
   param%eta1 = rparam_type(val=0.0,str='eta1')
   param%eta2 = rparam_type(val=0.0,str='eta2')
   param%asym = rparam_type(val=0.0,str='asym')

   end function init_psvoigt_param

!------------------------------------------------------------------------------------

   subroutine psvoigt(par,ref,xcount,fs,scal,ycount,nwave,rapporto_int,ier)
   !USE progtype
   USE arrayutil
   USE profile_function
   USE reflection_type_util
!
   type(pvoigt_function_type), intent(in)             :: par
   type(reflection_type), dimension(:), intent(inout) :: ref
   real, dimension(:), intent(in)                     :: xcount
   real, dimension(:), intent(in)                     :: fs    
   real, intent(in)                                   :: scal
   real, dimension(:), intent(out)                    :: ycount
   real, intent(in)                                   :: rapporto_int
   integer, intent(in)                                :: nwave
   integer, intent(out)                               :: ier
   real, dimension(2)                                 :: rap
   integer             :: j,nnn
   integer             :: iniz,ifin
   real                :: eta,eta0,eta1,eta2
   real                :: fu,fv,fw
   real                :: uvw2,uvwqg
   real                :: asym
   real                :: tthdeg
   real                :: rdeg,tdeg,tdeg2
   real                :: cgaus,cgaus2,cloren,cloren2
!corr   real                :: gausconst,gausconst2
!corr   real                :: constloren
   real                :: ffx
   real                :: deltath,deltath2
   real                :: pvoigt,asterm
   integer             :: negfw
   integer             :: kal
!
   ycount(:) = 0.0
!
   call init_calc(size(ref))
!
   eta0 = par%eta0%val
   eta1 = par%eta1%val
   eta2 = par%eta2%val
   fu = par%u%val
   fv = par%v%val
   fw = par%w%val
   asym = par%asym%val
!corr   gausconst2 = 4.0*alog(2.0)
!corr   gausconst = sqrt(gausconst2/pi)
!corr   constloren=2.0/pi
   negfw = 0
   ier = 0
   rap = (/1.0,rapporto_int/)
   LOOP_KALPHA : do kal=1,nwave
      LOOP_RIFLESSI : do j=1,size(ref)
         tthdeg = ref(j)%tthd  
         rdeg = 0.5 * tthdeg * dtor
         tdeg = sin(rdeg) / cos(rdeg)
         tdeg2 = tdeg * tdeg
!
         eta = eta0 + eta1*tdeg + eta2*tdeg2
         if (eta < 0 .or. eta > 1) then
             negfw = negfw + 1
         endif
         etav(j) = eta
         uvwqg = fu*tdeg2 + fv*tdeg + fw
         if (uvwqg < 0.0001) then
             uvwqg = 0.0001
             negfw = negfw + 1
         endif
         ruvw(j) = 1./uvwqg
         ref(j)%fwhm = sqrt(uvwqg)
         cgaus =   GAUSCONST / ref(j)%fwhm
         cloren = CONSTLOREN / ref(j)%fwhm
         !cgaus2 =   GAUSCONST2 / uvwqg
         !cloren2 = 4.0 / uvwqg
         cgaus2 =   GAUSCONST2 * ruvw(j)
         cloren2 = 4.0 * ruvw(j)

!corr         phkl(j,1)=cgaus
!corr         phkl(j,2)=cgaus2
!corr         phkl(j,3)=cloren
!corr         phkl(j,4)=cloren2
!
         uvw2 =  ref(j)%fwhm * ref(j)%pk * rapIcalc(j)
!
         iniz = clocate(xcount,tthdeg-uvw2)
         ifin = clocate(xcount,tthdeg+uvw2)
!
         ffx = ref(j)%m*fs(j)**2*ref(j)%po*ref(j)%lp*ref(j)%as*rap(kal)
!
         LOOP_CONTEGGI : do nnn=iniz,ifin
            deltath=xcount(nnn)-tthdeg
            deltath2=deltath**2
            pvoigt = eta*cloren/(1.0+cloren2*deltath2) + (1.0-eta)*cgaus*exp(-cgaus2*deltath2)
            asterm=1.0-(asym*sign(1.0,deltath)*deltath2)/tdeg
            ycount(nnn)=ycount(nnn)+scal*ffx*pvoigt*asterm
         enddo LOOP_CONTEGGI
      enddo LOOP_RIFLESSI
   enddo LOOP_KALPHA
!!!!   if(negfw>0)write(0,*)'NWGFW=',negfw,real(negfw)/size(ref)
   if (real(negfw)/size(ref) > 0.4) ier = 1
!
   end subroutine psvoigt

!--------------------------------------------------------------------

   subroutine psvoigt_init(ref,xcount,par,maxlim,rapporto_int)
   USE reflection_type_util, only: reflection_type
   USE arrayutil
   USE profile_function
!
   type(reflection_type), dimension(:), intent(inout) :: ref
   real, dimension(:), intent(in)                     :: xcount
   type(pvoigt_function_type), intent(in)             :: par
   integer, intent(out)                               :: maxlim
   integer                                            :: j,nnn
   real, intent(in)                :: rapporto_int
   real                :: eta,eta0,eta1,eta2
   real                :: fu,fv,fw
   real                :: uvw2,uvwqg
   real                :: asym
   real                :: tthdeg
   real                :: rdeg,tdeg,tdeg2
   real                :: cgaus,cgaus2,cloren,cloren2
!corr   real                :: gausconst,gausconst2
!corr   real                :: constloren
   real                :: deltath,deltath2
   real                :: pvoigt,asterm
   real, dimension(2)  :: rap
!
   if (.not.allocated(profi)) allocate(profi(size(ref))) 
!
   call init_calc(size(ref))
!
   eta0 = par%eta0%val
   eta1 = par%eta1%val
   eta2 = par%eta2%val
   fu = par%u%val
   fv = par%v%val
   fw = par%w%val
   asym = par%asym%val
!corr   gausconst2 = 4.0*alog(2.0)
!corr   gausconst = sqrt(gausconst2/pi)
!corr   constloren=2.0/pi
   rap = (/1.0,rapporto_int/)
   LOOP_RIFLESSI : do j=1,size(ref)
      tthdeg = ref(j)%tthd
      rdeg = 0.5 * tthdeg * dtor
      tdeg = sin(rdeg) / cos(rdeg)
      tdeg2 = tdeg * tdeg
!
      eta = eta0 + eta1*tdeg + eta2*tdeg2
      etav(j) = eta
      uvwqg = fu*tdeg2 + fv*tdeg + fw
      if (uvwqg < 0.0001)  uvwqg = 0.0001
!corr      ref(j)%fwhm = sqrt(uvwqg)
!corr      cgaus =   GAUSCONST/ref(j)%fwhm
!corr      cgaus2 =   GAUSCONST2/uvwqg
!corr      cloren = CONSTLOREN / ref(j)%fwhm
!corr  cloren2 = 4.0 / uvwqg
      ruvw(j) = 1./uvwqg
      ref(j)%fwhm = sqrt(uvwqg)
      cgaus =   GAUSCONST / ref(j)%fwhm
      cloren = CONSTLOREN / ref(j)%fwhm
      !cgaus2 =   GAUSCONST2 / uvwqg
      !cloren2 = 4.0 / uvwqg
      cgaus2 =   GAUSCONST2 * ruvw(j)
      cloren2 = 4.0 * ruvw(j)
!
      uvw2 =  ref(j)%fwhm * ref(j)%pk * rapIcalc(j)
!
      profi(j)%iniz = clocate(xcount,tthdeg-uvw2)
      profi(j)%ifin = clocate(xcount,tthdeg+uvw2)
      allocate(profi(j)%func(profi(j)%ifin-profi(j)%iniz+1))
      profi(j)%func = 0.0  ! forse non serve
!
      LOOP_CONTEGGI : do nnn=profi(j)%iniz,profi(j)%ifin
         deltath=xcount(nnn)-tthdeg
         deltath2=deltath**2
         pvoigt = eta*cloren/(1.0+cloren2*deltath2) + (1.0-eta)*cgaus*exp(-cgaus2*deltath2)
         asterm=1.0-(asym*sign(1.0,deltath)*deltath2)/tdeg
         profi(j)%func(nnn-profi(j)%iniz+1) = pvoigt*asterm
      enddo LOOP_CONTEGGI
      !profi(j)%fconst = ref(j)%m*ref(j)%po*ref(j)%lp
      profi(j)%fconst = ref(j)%m*ref(j)%lp
   enddo LOOP_RIFLESSI
   maxlim = maxval(profi%ifin)
!
   end subroutine psvoigt_init

!-------------------------------------------------------------------------------------
  
   subroutine derpsvoi(par,inizr,ifinr,xcount,ref,cell,derFct,derOP,deryz,dfw,derp,dercc,derG,  &
                       rapporto_int,scalr,rinfo,nwave,gpo,neutro,nsyn,wave,jsys)
   USE progtype, only: refine_info_type
   USE arrayutil
   USE profile_function
   USE reflection_type_util
!
   type(pvoigt_function_type), intent(in)          :: par
   integer, intent(in)                             :: inizr,ifinr
   real, dimension(:), intent(in)                  :: xcount
   type(reflection_type), dimension(:), intent(in) :: ref
   real, dimension(6), intent(in)                  :: cell
   real, dimension(:), intent(in)                  :: derFct
   real, dimension(:), intent(in)                  :: derOP
   real, dimension(:), intent(out)                 :: deryz
   real, dimension(:,:), intent(out)               :: dfw
   real, dimension(:,:), intent(out)               :: derp
   real, dimension(:,:), intent(out)               :: dercc
   real, dimension(:), intent(out)                 :: derG
   real, intent(in)                                :: rapporto_int
   type(rparam_type), intent(in)                   :: scalr
   type(refine_info_type), intent(in)              :: rinfo
   integer, intent(in)                             :: nwave
   type(rparam_type), intent(in)                   :: gpo
   integer, intent(in)                             :: neutro, nsyn, jsys
   real, intent(in)                                :: wave
   integer             :: j,k,kal,nnn
   integer             :: iniz,ifin
   logical             :: iderc
   real                :: asym,scal
   real                :: eta0,eta1,eta2,eta
   real                :: tthdeg
   real                :: rdeg,tdeg,tdeg2
   real                :: uvwg,uvwqg,uvw2
   real                :: pesolp,absoc
   real                :: derft,dlorp
   real                :: ffnop,ffx,ffb1
   real                :: deltath,deltath2,cc
   real                :: asterm,pvoigt,pgau,plor
   real                :: deruvw,dereta,derasth
   real                :: derpth,deryop
   real, dimension(6)  :: dercells
   real                :: derfunczero,deraszero
   real, dimension(2)  :: rap
   real, dimension(3)  :: vetder,gradprof
   real                :: constg
   real                :: fu,fv    !!!!!,fw
   real                :: secth2
   real                :: kpol
   real                :: cgaus,cgaus2,cloren,cloren2
!
   if (neutro == 1 .or. nsyn == 1) then
       kpol = 0.0
   else
       kpol = 0.8
   endif
   deryz = 0.0
   dfw   = 0.0
   derp  = 0.0
   dercc = 0.0
   derG(:)  = 0.0
!
   vetder(1) = 1.0
   rap(1) = 1.0
   rap(2) = rapporto_int
   eta0 = par%eta0%val
   eta1 = par%eta1%val
   eta2 = par%eta2%val
   fu = par%u%val
   fv = par%v%val
   asym = par%asym%val

   scal = scalr%val
   iderc = rinfo%refine_cell 
   constg = 4.0*alog(2.0)
   LOOP_KALPHA : do kal=1,nwave
      LOOP_RIFLESSI : do j=inizr,ifinr
         tthdeg = ref(j)%tthd  
         derft = derFct(j)  
         eta = etav(j)
         rdeg = 0.5 * tthdeg * dtor
         tdeg = sin(rdeg) / cos(rdeg)
         tdeg2 = tdeg * tdeg
         secth2 = 1.0 / cos(rdeg)**2
         uvwg = ref(j)%fwhm
         uvwqg = uvwg*uvwg
         !cgaus2 =   GAUSCONST2 / uvwqg
         !cgaus2 =  phkl(j,2)
         cgaus =   GAUSCONST / ref(j)%fwhm
         cloren = CONSTLOREN / ref(j)%fwhm
         cgaus2 =   GAUSCONST2 * ruvw(j)
         cloren2 = 4.0 * ruvw(j)

         uvw2 = ref(j)%fwhm * ref(j)%pk * rapIcalc(j)
!
         iniz = clocate(xcount,tthdeg-uvw2)
         ifin = clocate(xcount,tthdeg+uvw2)
!
         if(iderc)then
            dlorp = derivataLP(ref(j)%tthd,kpol)
            !dlorp = 0.0 
            dercells = scal*derteta_risp_celldir(jsys,cell,ref(j)%hkl,ref(j)%tthd,wave)   !(der. di theta risp. alla cella) * scala
            vetder(2) = secth2 * (eta1 + 2.0*eta2*tdeg)               !der. eta rispetto theta del riflesso
            vetder(3) = secth2 * (fv + 2.0*fu*tdeg)*0.5/uvwg        !der. FWHM rispetto theta del riflesso
         endif
         pesolp = ref(j)%lp
         absoc = ref(j)%as
         ffnop = pesolp*ref(j)%m*(ref(j)%fc**2)*absoc
         ffx = ffnop*ref(j)%po
         ffb1 = rap(kal)*pesolp*ref(j)%m*ref(j)%po*absoc
         LOOP_CONTEGGI : do nnn=iniz,ifin
            deltath=xcount(nnn)-tthdeg
            deltath2=deltath**2
            !cc=(1.0+phkl(j,4)*deltath2)
            cc=(1.0+cloren2*deltath2)
            !plor= phkl(j,3)/cc
            !pgau= phkl(j,1)*exp(-phkl(j,2)*deltath2)
            plor= cloren/cc
            pgau= cgaus*exp(-cgaus2*deltath2)
            pvoigt = eta*plor + (1.0-eta)*pgau
            asterm=1.0-(asym*sign(1.0,deltath)*deltath2)/tdeg
!
!           Derivative of 2theta corrected
            !deraszero = 4.0*asym*sign(1.0,deltath)*deltath/tdeg                                  ! der. asym. rispetto zero
            !derfunczero = 4.0*deltath*(cloren2*plor*eta/cc + cgaus2*(1.0-eta)*pgau) ! der. pvoigt rispetto zero
            deraszero = -2.0*asym*sign(1.0,deltath)*deltath/tdeg                     ! der. asym. rispetto 2theta corrected
            derfunczero = -2.0*deltath*(cloren2*plor*eta/cc + cgaus2*(1.0-eta)*pgau) ! der. pvoigt rispetto 2theta corrected
            deryz(nnn) = deryz(nnn) + scal * ffx * (derfunczero*asterm + deraszero*pvoigt)
!
!           Derivata Pseudo-Voigt rispetto eta
!
            gradprof(2) = plor - pgau
!
!           Derivata Pseudo-Voigt rispetto FWHM
!
            gradprof(3) = ( eta*plor*(8.0*deltath2-cc*uvwqg)/cc + (eta-1.0)*pgau*(uvwqg-2.0*constg*deltath2) ) / uvwg**3
!
            if (rinfo%refine_profile) then
              deruvw = scal * ffx * asterm * gradprof(3) / (2.0*uvwg)   !gradprof(3)/(2.0*uvwg) = der. rispetto FWHM**2
              dfw(nnn,1)=dfw(nnn,1) + deruvw * tdeg2    !der. rispetto U
              dfw(nnn,2)=dfw(nnn,2) + deruvw * tdeg     !der. rispetto V
              dfw(nnn,3)=dfw(nnn,3) + deruvw            !der. rispetto W
!
              derp(nnn,1)=derp(nnn,1) - scal*ffx*pvoigt*sign(1.0,deltath)*deltath2/tdeg !der. rispetto asym
!
              dereta = scal * ffx * asterm * gradprof(2)
              derp(nnn,2)=derp(nnn,2) + dereta            !der. rispetto eta0
              derp(nnn,3)=derp(nnn,3) + dereta * tdeg     !der. rispetto eta1
              derp(nnn,4)=derp(nnn,4) + dereta * tdeg2    !der. rispetto eta2
            endif
            if (iderc) then
               derasth=asym*sign(1.0,deltath)*deltath*(rtod*4.0/tdeg + deltath/(sin(rdeg)**2)) !der.asym risp. theta
!
!              Derivata Psvoig rispetto theta del riflesso
!
               gradprof(1) = rtod*4.0*deltath*(eta*4.0*plor/cc + (1-eta)*constg*pgau)/uvwqg
!
               derpth = DOT_PRODUCT(gradprof,vetder)       !derivata rispetto theta del riflesso
               do k=1,6
                   dercc(nnn,k)=dercc(nnn,k)+dercells(k)*            &
                        (ffx*(derpth*asterm+derasth*pvoigt)+(ffb1*derft+ffx/pesolp*dlorp)*pvoigt*asterm)
               enddo
            endif
!
            if (gpo%rcod > 0) then
                deryop = scal * ffnop * pvoigt * asterm
                derG(nnn)=derG(nnn) + deryop*derOP(j)
            endif
         enddo LOOP_CONTEGGI
      enddo LOOP_RIFLESSI
   enddo LOOP_KALPHA
!
   end subroutine derpsvoi

!-----------------------------------------------------------------------

   subroutine derpsvoia(ref,par,inizr,ifinr,xcount,der,derc,nwave,rapporto_int,scalr)
   USE progtype, only : deratom
   USE reflection_type_util, only : reflection_type
   USE arrayutil
   USE profile_function
!
   type(reflection_type), dimension(:), intent(in) :: ref
   type(pvoigt_function_type), intent(in)          :: par
   integer, intent(in)                             :: inizr,ifinr
   real, dimension(:), intent(in)                  :: xcount
   type(deratom), dimension(:,:), intent(in)       :: der
   type(deratom), dimension(:,:), intent(out)      :: derc
   integer, intent(in)                             :: nwave
   real, intent(in)                                :: rapporto_int
   type(rparam_type), intent(in)                   :: scalr
   integer                                         :: j,jj,kal,nnn
   integer                                         :: iniz,ifin
   real                                            :: tthdeg
   real                                            :: rdeg,tdeg
   real, dimension(2)                              :: rap
   real                                            :: asym,scal,eta,uvw2
   real                                            :: deltath,deltath2
   real                                            :: asterm,pvoigt
   real                                            :: ffb1,dydI
   real                                            :: cgaus1,cgaus2,cloren1,cloren2
   integer                                         :: nat
!
   derc = deratom(0.0,0.0,0.0)
   nat = size(derc,2)
!
   rap(1) = 1.0
   rap(2) = rapporto_int
   asym = par%asym%val
   scal = scalr%val
!corr   constg1 = 4.0*alog(2.0)
!corr   constg2 = sqrt(constg1/pi)
!corr   CONSTLOREN  = 2.0/pi
   LOOP_KALPHA : do kal=1,nwave
      LOOP_RIFLESSI : do j=inizr,ifinr
         tthdeg = ref(j)%tthd
         rdeg = 0.5 * tthdeg * dtor
         tdeg = sin(rdeg) / cos(rdeg)
!
!corr         fwhm = ref(j)%fwhm
         !cgaus1 =   constg1/fwhm**2
         !cgaus2 =   constg2/fwhm
         cgaus2 =   GAUSCONST/ref(j)%fwhm
         cloren1 = CONSTLOREN / ref(j)%fwhm
         !cloren2 = 4.0 / fwhm**2
         cloren2 = 4.0 * ruvw(j)
         cgaus1 =   GAUSCONST2 * ruvw(j)
!
         eta = etav(j)
         uvw2 = ref(j)%fwhm * ref(j)%pk * rapIcalc(j)
!
         iniz = clocate(xcount,tthdeg-uvw2)
         ifin = clocate(xcount,tthdeg+uvw2)
!
         !!!!!opT = sum(frac*oprcorr(j,:))
         !!!!!opT = oprcorr(j,1)
         ffb1 = rap(kal)*ref(j)%lp*ref(j)%m*ref(j)%po*ref(j)%as
         LOOP_CONTEGGI : do nnn=iniz,ifin
            deltath=xcount(nnn)-tthdeg
            deltath2=deltath**2
            pvoigt = eta*cloren1/(1.0+cloren2*deltath2) + (1.0-eta)*cgaus2*exp(-cgaus1*deltath2)
            asterm=1.0-(asym*sign(1.0,deltath)*deltath2)/tdeg
!
!           calcolo derivate param. strutt. per atomo
!
            dydI = scal * ffb1 * pvoigt * asterm
            do jj=1,nat
               derc(nnn,jj)%co = derc(nnn,jj)%co + dydI * der(j,jj)%co
               derc(nnn,jj)%b = derc(nnn,jj)%b + dydI * der(j,jj)%b
               derc(nnn,jj)%occ = derc(nnn,jj)%occ + dydI * der(j,jj)%occ
            enddo
!
         enddo LOOP_CONTEGGI
      enddo LOOP_RIFLESSI
   enddo LOOP_KALPHA
!
   end subroutine derpsvoia

!--------------------------------------------------------------------

   subroutine foleb_psvoi(ref,inizr,ifinr,Icalc,Iobs,pvoi,scalr,yoss,ytot,ttetaczero,back)
   USE arrayutil
   USE profile_function, only: rapIcalc
   USE reflection_type_util, only: reflection_type
!
   type(reflection_type), dimension(:), intent(in) :: ref
   integer, intent(in)                             :: inizr,ifinr
   real, dimension(:), intent(in)                  :: Icalc
   real, dimension(:), intent(out)                 :: Iobs
   type(pvoigt_function_type), intent(in)          :: pvoi
   type(rparam_type), intent(in)                   :: scalr
   real, dimension(:), intent(in)                  :: yoss,ytot,ttetaczero,back
   integer                                         :: iniz,ifin
   real                                            :: asym,eta,scal
   integer                                         :: j,nnn
   real                                            :: rdeg,tdeg
   real                                            :: tthdeg,deltath,deltath2
   real                                            :: uvw2
   real                                            :: asterm,yci,diffo,ycalc
   real                                            :: plor,pgau,pvoigt
   real                                            :: sumy
   integer                                         :: nneg
   real                                            :: epsmch = epsilon(1.0)
   real                                            :: cgaus,cgaus2,cloren,cloren2
!
   nneg = 0
   asym = pvoi%asym%val
   scal = scalr%val
   LOOP_RIFLESSI : do j=inizr,ifinr
      sumy = 0.0
      tthdeg = ref(j)%tthd
!
      rdeg = 0.5 * tthdeg * dtor
      tdeg = sin (rdeg) / cos (rdeg)
      eta = etav(j)
      uvw2 = ref(j)%fwhm * ref(j)%pk * rapIcalc(j)
      cgaus =   GAUSCONST / ref(j)%fwhm
      cgaus2 =   GAUSCONST2 * ruvw(j)
      cloren = CONSTLOREN / ref(j)%fwhm
      cloren2 = 4.0 * ruvw(j)
!
      iniz = clocate(ttetaczero,tthdeg-uvw2)
      ifin = clocate(ttetaczero,tthdeg+uvw2)
!
      LOOP_CONTEGGI : do nnn=iniz,ifin
         deltath=ttetaczero(nnn)-tthdeg
         deltath2=deltath**2
         !plor= phkl(j,3)/(1.0+phkl(j,4)*deltath2)
         plor= cloren/(1.0+cloren2*deltath2)
         !pgau= phkl(j,1)*exp(-phkl(j,2)*deltath2)
         pgau= cgaus*exp(-cgaus2*deltath2)
         pvoigt = eta*plor + (1.0-eta)*pgau
         asterm=1.0-(asym*sign(1.0,deltath)*deltath2)/tdeg
         yci = scal * pvoigt * asterm
         if (ytot(nnn) > epsmch) then
            diffo = yoss(nnn)-back(nnn)
            if (diffo.lt.0.0) then
                cycle
            endif
            ycalc = diffo*yci/ytot(nnn)
         else
            cycle
         endif
         sumy = sumy + ycalc
      enddo LOOP_CONTEGGI
!
      Iobs(j) = Icalc(j) * sumy
!!!!      if (Iobs(j) < 0.0) then
      if (Iobs(j) < epsmch) then
          Iobs(j) = 0.0
          nneg = nneg + 1
      endif
!
   enddo LOOP_RIFLESSI
   !if (nneg > 0) write(6,*)'IOBS NEGATIVI n=',nneg
!
   end subroutine foleb_psvoi
  !------------------------------------------------------------------------------------------

end module psvoigtf
