!  This file is an example of the Corana et al. simulated annealing algorithm
!  for multimodal and robust optimization as implemented and modified by
!  Goffe, Ferrier and Rogers. This version in Fortran 90 has been prepared by Corrado Cuocci
!  using a previous version of Alan Miller. 
MODULE sa_type
type, public        :: SimAnn_Conditions_type
  real              :: t_ini       ! Initial temperature
  real              :: temp_fact   ! Temperature reduction factor
  real              :: eps         ! Error tolerance for termination
  integer           :: initconfig  ! Flag determining if the first configuration is random or read
  integer           :: nt          ! Number of iterations before temperature reduction
  integer           :: num_temps   ! Maximum number of temperatures in SA
  integer           :: ns          ! After ns*npar stp is updated
  integer           :: neps        ! Number of final function values used to decide upon termination
  logical           :: autoti      ! calcolo automatico t_ini?
  logical           :: autont      ! calcolo automatico nt?
  integer           :: rand        ! randomizzazione iniziale
end type SimAnn_Conditions_type

END MODULE sa_type

MODULE SA_INTERFACE
   INTERFACE
      subroutine fcnprint(gopti,popt,xval,fopt,temp,nftot,nmtemp)
      USE progtype
      implicit none
      type(opti_info_type), intent(in) :: gopti
      character(len=*), intent(in)   :: popt
      real, dimension(:), intent(in) :: xval
      real, intent(in)               :: fopt
      real, intent(in)               :: temp
      integer, intent(in)            :: nftot
      integer, intent(in)            :: nmtemp
      end subroutine fcnprint

      subroutine gestione_eventi(kevent)
      implicit none
      integer, intent(out) :: kevent
      end subroutine gestione_eventi
   END INTERFACE
END MODULE SA_INTERFACE

MODULE simulated_anneal
   USE sa_type
   USE progtype
   implicit none
   logical, public            :: err_san = .false.    ! error flag      
   character(len=150), public :: err_mess_san = " "   ! error message
!
!  variabili per gestione messaggi
   integer, dimension(20)     :: vmess
   integer, parameter         :: LINE_RUN=2, LINE_TEMP=3, LINE_MOVES=4, LINE_CF=5, LINE_RATIO=6, &
                              LINE_TIME=7, LINE_CFMIN=8, LINE_CFBUMP=9, LINE_CFRES=10

 CONTAINS

   subroutine sa(par,sc,ginfo,temp,xopt,fopt,nacc,nfcnev,nobds,kevent,func,funcs,freset)
   
!  Version: 4.0. Date: 25/04/2007.
!  
!  This routine implements the continuous simulated annealing global
!  optimization algorithm described in Corana et al.'s article "Minimizing
!  Multimodal Functions of Continuous Variables with the "Simulated Annealing"
!  Algorithm" in the September 1987 (vol. 13, no. 3, pp. 262-280) issue of
!  the ACM Transactions on Mathematical Software.   
!corr   USE calctime
   USE SA_INTERFACE
   USE lsqalgo, only:steepest_descent
   USE stat
   USE rand_mod
   USE messagemod
   USE strutil
   INTERFACE
      real function func(xpar,hpar)
      real, intent(in), dimension(:) :: xpar
      integer, intent(in)            :: hpar
      end function func

      subroutine funcs(xp,fval,grad,icod)
      real, dimension(:), intent(in)  :: xp
      real, intent(out)               :: fval
      real, dimension(:), intent(out), optional :: grad
      integer, intent(in)             :: icod
      end subroutine funcs
   END INTERFACE

   type(param_type), dimension(:), intent(inout) :: par    !The variables to be optimized
   type(SimAnn_Conditions_type), intent(inout)   :: sc     !Contains conditions for running Simulated Annealing
   type(opti_info_type), intent(in)              :: ginfo
   real, intent(out)                             :: temp   !Temperature
   real, dimension(:), intent(out)               :: xopt   !The variables that optimize the function
   real, intent(out)                             :: fopt   !The optimal value of the function
   integer, intent(out)                          :: nacc   !The number of accepted function evaluations
   integer, intent(inout)                        :: nfcnev !The total number of function evaluations
   integer, intent(out)                          :: nobds  !The total number of trial function evaluations out of bounds of LB and UB
   integer, intent(out)                          :: kevent !evento di terminazione dalla grafica
   real, intent(in)                              :: freset !azzera il numero di cicli
   real, dimension(size(par))                    :: cv     !Vector that controls the step length adjustment.
   integer                                       :: npar      
   real                                          :: random
   logical                                       :: print_graph,print_cf
   real, dimension(size(par))                    :: trialval
   real                                          ::  fval, ftrial,prob, ratio, fstar(sc%neps)
   integer                                       :: nup, ndown, nrej, nnew 
   integer                                       :: lnobds, h, i, j, m, nacp(size(par))
   logical                                       :: quit
   real                                          :: range
   integer                                       :: ntp
   integer                                       :: nnopt
!test   integer                                       :: nacc_per_temp, ntrial_per_temp
!test   real :: temp_max, temp_min
!corr   real                                          :: cffin,cfinit,betak
!corr   integer                                       :: niter
!   
   call sa_check(sc)
!
!  Number of parameters
   npar = size(par)   
!   
!  Set initial values.
   kevent = 0
   nacc = 0
   nobds = 0
!test   ntrial_per_temp = sc%nt*sc%ns*npar
!test   temp_max = sc%t_ini
!test   temp_min = 0.01
!
!  freset > 0: non azzerare il numero di cicli e aggiorna grafica solo se fopt < freset
   if (freset < 0) nfcnev = 0
!
   nnopt = 0
   cv(:) = 2.0   
   print_graph = index(ginfo%popt,'P') /= 0 .or. index(ginfo%popt,'M') /= 0
   print_cf = index(ginfo%popt,'V') /= 0

   xopt(:) = par(:)%val
   nacp(:) = 0
   temp = sc%t_ini   
   fval = func(par%val,0)
   fopt = fval
   call init_convergence(fopt,fstar)
   
   if (print_cf) then
       if (freset < 0 .or. (freset > 0 .and. fopt < freset)) then
           call fcnprint(ginfo,'V',xopt,fopt,temp,nfcnev,0)
       endif
   endif
!   
!  Start the main loop. Note that it terminates if (i) the algorithm
!  succesfully optimizes the function or (ii) there are too many
!  function evaluations (more than MAXEVL).
   loop_temp: do ntp=1,sc%num_temps               ! global do for changing temperature
      if (print_graph) then
          call write_message('Run '//trim(i_to_s(ginfo%nrunc))//'/'//i_to_s(ginfo%nrun),pos=vmess(LINE_RUN))
          call write_message('Temperature: ',rnum=temp,pos=vmess(LINE_TEMP))
      endif
      nup = 0
      nrej = 0
      nnew = 0
      ndown = 0
      lnobds = 0
!test      nacc_per_temp = 0
!corr      cfinit = fopt
!corr      niter = 0
      
      loop_moves: do m=1,sc%nt                      ! loop over Montecarlo moves at fixed temperature          
         do j = 1, sc%ns                ! After this loop sc%stp is updated
            loop_par: do h=1,npar       ! loop on parameters
!
!corr               niter = niter + 1
!
!              Gestione eventi di interazione con la grafica
               call gestione_eventi(kevent)
               select case (kevent)
                   case (1,2)    ! stop o skip
                      exit loop_moves
               end select
!            
!              Generate XP, the trial value of X. Note use of VM to choose XP.
               do i = 1, npar
                  if (i == h) then
                      call random_number(random)
                      trialval(i) = par(i)%val + (random*2. - 1.) * par(i)%stp
                  else
                      trialval(i) = par(i)%val
                  endif
!                 
!                 Check for parameters out of bounds
                  range = par(i)%high - par(i)%low
                  if (par(i)%bound == 0) then       ! Fixed boundaries
                      if((trialval(i) < par(i)%low) .OR. (trialval(i) > par(i)%high)) then
                         call random_number(random)     
                         trialval(i) = par(i)%low + range*random
                      endif
                      lnobds = lnobds + 1
                      nobds = nobds + 1            
                  else                              ! Periodic boundaries
                      if(trialval(i) < par(i)%low) then
                         trialval(i) = trialval(i) + range
                         lnobds = lnobds + 1
                         nobds = nobds + 1                  
                      endif   
                      if(trialval(i) > par(i)%high) then
                         trialval(i) = trialval(i) - range 
                         lnobds = lnobds + 1
                         nobds = nobds + 1                  
                      endif             
                  endif
               enddo
!           
!              Evaluate the function with the trial point trialval and return as FP.
               ftrial = func(trialval, h)
               nfcnev = nfcnev + 1
               !if (print_graph) call fcnprint('B',xopt,fopt,temp,nfcnev,nup+ndown+nrej+1)
               if (print_graph) call fcnprint(ginfo,'B',xopt,fopt,temp,nfcnev,nfcnev)
               if (ftrial <= fval) then
!              
!                  Accept the new point if the function value decreases      
                   par%val = trialval  
                   fval = ftrial
                   nacc = nacc + 1
!test                   nacc_per_temp = nacc_per_temp + 1
                   nacp(h) = nacp(h) + 1
                   nup = nup + 1
                   call sa_savepar(h)
!                
!                  If greater than any other point, record as new optimum.
                   if (ftrial < fopt) then
!
!                      minimizza
                       !if (ftrial/fopt < 0.9) then
                       !    write(6,*)'steep prima:',ftrial
                       !    call steepest_descent(funcs,trialval,ftrial,6,0.001)
                       !    write(6,*)'steep dopo:',ftrial
                       !    fval = ftrial
                       !    par%val = trialval  
                       !endif
!
                       xopt(1:npar) = trialval
                       fopt = ftrial
                       nnew = nnew + 1
                       nnopt = nnopt + 1
!
                       if (freset < 0 .or. (freset > 0 .and. fopt < freset)) then
                           call fcnprint(ginfo,ginfo%popt,xopt,fopt,temp,nfcnev,0)
                       endif
                   endif
               else
!              
!                  If the point is higher, use the Metropolis test to decide on acceptance or rejection      
                   prob = exprep(-(ftrial - fval)/temp)
                   call random_number(random)
                   if (random < prob) then
                       par%val = trialval
                       fval = ftrial
                       nacc = nacc + 1
!test                       nacc_per_temp = nacc_per_temp + 1
                       nacp(h) = nacp(h) + 1
                       ndown = ndown + 1
                       call sa_savepar(h)
                   else
                       nrej = nrej + 1
                       call sa_restorepar(h)
                   endif
               endif
              
            enddo loop_par    ! end loop on parameters
         enddo                ! end loop do j = 1, sc%ns
!     
!        Adjust STP so that approximately half of all evaluations are accepted.
!        Corana's algorithm.
         do i = 1, npar
           range = par(i)%high - par(i)%low
           ratio = real(nacp(i)) /real(sc%ns)
           if (ratio > 0.6) then
               par(i)%stp = par(i)%stp*(1. + cv(i)*(ratio - 0.6)/0.4)
           elseif (ratio < 0.4) then
               par(i)%stp = par(i)%stp/(1. + cv(i)*((0.4 - ratio)/0.4))
           endif
           if (par(i)%stp > range) then
               par(i)%stp = range
           endif
         enddo
                 
         nacp(:) = 0
        
      enddo loop_moves   ! end loop over Montecarlo moves at fixed temperature          
!
      !cffin = fopt
      !betak = max(0.90/(1.0+0.2*real(nrej)/(nrej+ndown+nup)),1-cffin/cfinit)
      !write(6,*)'betak=',betak,nrej,nrej+ndown+nup,niter
    !  call avevar(cfvalue(:),cfave,cfvar)
    !  cfsig = sqrt(cfvar) 
    !  write(6,*)'sig=',cfsig,' ave=',cfave,' max=',maxval(cfvalue)
    !  betak=1.0/(1+0.6*temp/(3*cfsig))
    !  write(6,*)'betak=',betak
!      
!corr      call timestep(1)
!      
!     Check termination criteria.
      call check_convergence(sc%eps,fval,fopt,fstar,quit)
!      
!     Terminate SA if appropriate.
      if (quit .or. kevent == 1 .or. kevent == 2) then
        par%val = xopt(1:npar)
        fval = func(par%val, 0)      
        return
      endif
!      
!     If termination criteria is not met, prepare for another loop.
      temp = sc%temp_fact*temp
!test      if (nacc_per_temp/real(ntrial_per_temp) > 0.30) then
!test          temp = sc%temp_fact*temp
!test      elseif (nacc_per_temp/real(ntrial_per_temp) < 0.10) then
!test          temp = temp/sc%temp_fact
!test      endif
!test      if (temp > temp_max) temp = temp_max
!test      if (temp < temp_min) temp = temp_min
!test             write(0,*)'NACC=',temp,temp_max,nacc_per_temp/real(ntrial_per_temp)
      !temp = betak*temp
!
      fval = fopt
      par(:)%val = xopt(:)
      fval = func(par%val, 0)      
   enddo loop_temp      ! end global do for changing temperature

   end subroutine sa

!---------------------------------------------------------------------------------------
   subroutine init_convergence(cfval,fstar)
   real, intent(in)                  :: cfval
   real, dimension(:), intent(inout) :: fstar
   integer                           :: i
!
   fstar(1) = cfval
   do i=2,size(fstar)
      fstar(i) = huge(1.0)
   enddo
!
   end subroutine init_convergence

!---------------------------------------------------------------------------------------

   subroutine check_convergence(eps,cfval,cfopt,fstar,quit)
   real, intent(in)                  :: eps
   real, intent(in)                  :: cfval,cfopt
   real, dimension(:), intent(inout) :: fstar
   logical, intent(out)              :: quit
   integer                           :: i
   integer                           :: nstar
!
   quit = .false.
   fstar(1) = cfval
   nstar = size(fstar)
   if ((fstar(1)-cfopt) <= eps) quit = .true.
   do i = 1, nstar
      if (abs(cfval - fstar(i)) > eps) quit = .false.
   enddo
   if (.not.quit) then
       do i = nstar, 2, -1
         fstar(i) = fstar(i-1)
       enddo
   endif
      !write(0,*)'fstar=',fstar(:)
!
   end subroutine check_convergence

!---------------------------------------------------------------------------------------

   subroutine sa_glob_old(par,sc,ginfo,temp,fopt,nfcnev,kevent,func,funcs)
   USE SA_INTERFACE
!corr   USE calctime
   USE lsqalgo
   USE rand_mod
   INTERFACE
      subroutine func(xpar,h,hpar)
      implicit none
      real, intent(in)    :: xpar(:)
      real, intent(out)   :: h
      integer, intent(in) :: hpar
      end subroutine func

      subroutine funcs(xp,fval,grad,icod)
      real, dimension(:), intent(in)            :: xp
      real, intent(out)                         :: fval
      real, dimension(:), intent(out), optional :: grad
      integer, intent(in)                       :: icod
      end subroutine funcs
   END INTERFACE
   type(SimAnn_Conditions_type), intent(inout)   :: sc     !Contains conditions for running Simulated Annealing
   type(opti_info_type), intent(in)              :: ginfo
   type(param_type), dimension(:), intent(inout) :: par    !The variables to be optimized
   real, intent(out)                             :: temp   !Temperature
   real, intent(out)                             :: fopt
   integer, intent(out)                          :: nfcnev
   integer, intent(out)                          :: kevent !evento di terminazione dalla grafica
   integer                                       :: npar
   logical                                       :: print_graph,print_cf
   real                                          :: fval,ftrial
   real, dimension(sc%neps)                      :: fstar
   integer                                       :: ntp,m,i,j
   real, dimension(size(par))                    :: xtrial,xopt,xscale,xrandom,cv,xgrad   !!!!,wei,sdir
   integer, dimension(size(par))                 :: nacp
   real                                          :: random
   real                                          :: range
   logical                                       :: quit  
   real                                          :: prob,ratio
   type(SimAnn_Conditions_type) :: scc
!!!!!      integer :: nfcnevv
   real, dimension(size(par))                    :: xoptt
   type(param_type), dimension(size(par)) :: parl
     real :: foptt
!!!!!       integer :: nobdss
!!!!!       real :: tfin
!!!!!       integer :: nacc
       integer :: nobds
!!!!!    real :: init_time,end_time
    logical :: localo=.false.
   real :: norm
   real :: foptold
!!!!!   real :: fmin
!
   call sa_check(sc)
!!!!!!!   call Init_Ran(sc%seed)
!
!  Number of parameters
   npar = size(par)   
!   
!  Set initial values.
   cv(:) = 2.0   
   nacp(:) = 0
   nfcnev = 0
   kevent = 0
   nobds = 0
   print_graph = index(ginfo%popt,'P') /= 0 .or. index(ginfo%popt,'M') /= 0
   print_cf = index(ginfo%popt,'V') /= 0
   temp = sc%t_ini   
   call func(par%val, fval,0)  ! calcolo iniziale di CF
   fopt = fval
   xopt(:) = par(:)%val
   xtrial(:) = xopt
   ftrial = fval
   call init_convergence(fopt,fstar)
   xscale(:) = 0.5
      xgrad(:) = 1.0
      norm = 1
  foptold = fopt
                       !call calc_gradient(xgrad,xopt,fopt,funcgrad)
!
   loop_temp: do ntp=1,sc%num_temps               ! global do for changing temperature
      !call timestep(1)
      loop_moves: do m=1,sc%nt              ! loop over Montecarlo moves at fixed temperature          
          !write(0,*)'m=',m,sc%nt
          !        call cpu_time(init_time)
         do j = 1, sc%ns
          !write(0,*)'j=',j,sc%ns
          call gestione_eventi(kevent)
          call random_number(xrandom)
          xtrial(:) = par(:)%val + (xrandom(:)*2.0 - 1.0) * par(:)%stp
          do i=1,npar
             range = par(i)%high - par(i)%low
             if (par(i)%bound == 0) then       ! Fixed boundaries
                 if((xtrial(i) < par(i)%low) .OR. (xtrial(i) > par(i)%high)) then
                    call random_number(random)     
                    xtrial(i) = par(i)%low + range*random
                    nobds = nobds + 1
                 endif
             else                              ! Periodic boundaries
                 if(xtrial(i) < par(i)%low) then
                    xtrial(i) = xtrial(i) + range
                    nobds = nobds + 1
                 endif   
                 if(xtrial(i) > par(i)%high) then
                    xtrial(i) = xtrial(i) - range 
                    nobds = nobds + 1
                 endif             
             endif
          enddo
!
          call func(xtrial, ftrial,0)
          nfcnev = nfcnev + 1
          if (print_graph) call fcnprint(ginfo,'B',xopt,fopt,temp,nfcnev,(m-1)*sc%ns+j)
          if (ftrial <= fval) then
!              
!             Accept the new point if the function value decreases      
              par(:)%val = xtrial(:)  
              fval = ftrial
              nacp(:) = nacp(:) + 1
!        
!             If greater than any other point, record as new optimum.
              if (ftrial < fopt) then
!
!                 minimizza
                  !if (ftrial/fopt < 0.9) then
                      write(6,*)'steep prima:',ftrial
                      call steepest_descent(funcs,xtrial,ftrial,6,0.001)
                      write(6,*)'steep dopo:',ftrial
                      fval = ftrial
                      par%val = xtrial  
                  !endif
!
                  xopt(:) = xtrial
                  foptold = fopt
                  fopt = ftrial
                  call fcnprint(ginfo,ginfo%popt,xopt,fopt,temp,nfcnev,0)
!
                  if (localo) then
                      scc = sc
                      scc%t_ini = temp
                      scc%num_temps = 1
                      parl(:) = par(:)
                      parl(:)%val = xtrial(:)
                      write(0,*)'switch to local anneal.',ftrial
                      write(6,*)'switch to local anneal.',ftrial
            !          call sa(parl, scc, tfin, xoptt, foptt, nacc, nfcnevv, nobdss, kevent,func,funcs)
                      write(0,*)'end to local anneal.',foptt
                      write(6,*)'end to local anneal.',foptt
                      if (foptt < fopt) then
                           write(0,*)'nuovo ottimo in local anneal.',fopt,foptt
                           write(6,*)'nuovo ottimo in local anneal.',fopt,foptt
                           xopt(:) = xoptt
                           fopt = foptt
                           call fcnprint(ginfo,ginfo%popt,xopt,fopt,temp,nfcnev,0)
                           par(:)%val = xopt(:)
                           fval = fopt
                      endif
                   endif
              endif
          else
              prob = exprep(-(ftrial - fval)/temp)
              call random_number(random)
              if (random < prob) then
                  par(:)%val = xtrial(:)
                  nacp(:) = nacp(:) + 1
                  fval = ftrial
              endif
          endif
        enddo
!     
!        Adjust STP so that approximately half of all evaluations are accepted.
!        Corana's algorithm.
         do i = 1, npar
           range = par(i)%high - par(i)%low
           ratio = real(nacp(i)) /real(sc%ns)
           if (ratio > 0.6) then
               par(i)%stp = par(i)%stp*(1. + cv(i)*(ratio - 0.6)/0.4)
           elseif (ratio < 0.4) then
               par(i)%stp = par(i)%stp/(1. + cv(i)*((0.4 - ratio)/0.4))
           endif
           if (par(i)%stp > range) then
               par(i)%stp = range
           endif
         enddo

         nacp(:) = 0

      enddo loop_moves   ! end loop over Montecarlo moves at fixed temperature          
!
      call check_convergence(sc%eps,fval,fopt,fstar,quit)
      fval = fopt
      par(:)%val = xopt(:)
      call func(par%val, fval,0)      
      if (quit .or. kevent == 1 .or. kevent == 2) exit
      temp = sc%temp_fact*temp
!
   enddo loop_temp      ! end global do for changing temperature

   !CONTAINS
!
!       subroutine safunc1(xp,fval,xgrad,icod)
!       USE lsqalgo
!!
!!      Funzione che per  icod = 1  valuta la funzione costo
!!                        icod = 2  valuta il gradiente
!       real, dimension(:), intent(in)  :: xp
!       real, intent(out)               :: fval
!       real, dimension(:), intent(out) :: xgrad
!       integer, intent(in)             :: icod
!!
!       select case(icod)
!          case (1)
!            call func(xp,fval,0)
!
!          case (2)
!            call calc_gradient(xgrad,xp,0.0,funcgrad,icod=2)
!
!       end select
!!
!       end subroutine safunc1
!
!       real function funcgrad(x)
!       real, dimension(:), intent(in) :: x
!       call func(x,funcgrad,0)
!       end function funcgrad
!
   end subroutine sa_glob_old

!---------------------------------------------------------------------------------------

   subroutine sa_glob(par,sc,ginfo,temp,fopt,nfcnev,kevent,func,funcg)
   USE SA_INTERFACE
!corr   USE calctime
   USE lsqalgo
   USE rand_mod
   INTERFACE
      subroutine func(xpar,h,hpar)
      implicit none
      real, intent(in)    :: xpar(:)
      real, intent(out)   :: h
      integer, intent(in) :: hpar
      end subroutine func

      real function funcg(x)
      real, dimension(:), intent(in) :: x
      end function funcg
   END INTERFACE
   type(SimAnn_Conditions_type), intent(inout)   :: sc     !Contains conditions for running Simulated Annealing
   type(opti_info_type), intent(in)              :: ginfo
   type(param_type), dimension(:), intent(inout) :: par    !The variables to be optimized
   real, intent(out)                             :: temp   !Temperature
   real, intent(out)                             :: fopt
   integer, intent(out)                          :: nfcnev
   integer, intent(out)                          :: kevent !evento di terminazione dalla grafica
   integer                                       :: npar
   logical                                       :: print_graph,print_cf
   real                                          :: fval,ftrial
   real, dimension(sc%neps)                      :: fstar
   integer                                       :: ntp,m,n
   real, dimension(size(par))                    :: xtrial,xopt,xscale,xrandom,cv,xgrad,sdir
!!!!   real, dimension(size(par))                    :: deltax
   real, dimension(size(par),2**size(par)-1)        :: svdir
   integer, dimension(size(par))                 :: nacp
   real                                          :: random
   logical                                       :: quit  
   real                                          :: prob   !!!!!,ratio
!!!!!!   type(SimAnn_Conditions_type) :: scc
!!!!!!     integer :: nfcnevv
!!!!!!  real, dimension(size(par))                    :: xoptt
!!!!!!   type(param_type), dimension(size(par)) :: parl
!!!!!!     real :: foptt
!!!!!!       integer :: nobdss
!!!!!!       real :: tfin
!!!!!!       integer :: nacc
       integer :: nobds
!!!!!!    real :: init_time,end_time
!!!!!!    logical :: localo=.false.
      logical :: newopt = .false.
   real :: norm
   real :: foptold
!!!!   real :: fmin
   integer :: ndir
   real, dimension(4) :: stepx =(/1.0,0.1,0.01,0.001/)
   integer :: nx
   logical :: globan 
!
   call sa_check(sc)
!!!!!!!   call Init_Ran(sc%seed)
!
!  Number of parameters
   npar = size(par)   
   ndir = 2**npar - 1
!   
!  Set initial values.
   cv(:) = 2.0   
   nacp(:) = 0
   nfcnev = 0
   kevent = 0
   nobds = 0
   print_graph = index(ginfo%popt,'P') /= 0 .or. index(ginfo%popt,'M') /= 0
   print_cf = index(ginfo%popt,'V') /= 0
   temp = sc%t_ini   
   call func(par%val, fval,0)  ! calcolo iniziale di CF
   fopt = fval
   xopt(:) = par(:)%val
   xtrial(:) = xopt
   ftrial = fval
   call init_convergence(fopt,fstar)
   xscale(:) = 0.5
      xgrad(:) = 1.0
      norm = 1
  foptold = fopt
      svdir(:,:) = 1.0
  globan = .true.
!
   loop_temp: do ntp=1,sc%num_temps               ! global do for changing temperature
      !call timestep(1)
      loop_moves: do m=1,sc%nt              ! loop over Montecarlo moves at fixed temperature          
          call gestione_eventi(kevent)
          newopt = .false.
          !if (globan) then
!
!             Calcola il gradiente
              call calc_gradient(xgrad,xopt,fopt,funcg,icod=2)
              norm = sqrt(dot_product(xgrad,xgrad))   ! calcola la norma
!
!             Genera la prima direzione discendente
              call random_number(xrandom)
              sdir(:) = - (par(:)%high - par(:)%low)*xrandom(:)*(xgrad(:)/norm)*(foptold/fopt)
!
!             Genera le altre direzioni discendenti
              call genera_direzioni(xgrad,sdir,svdir)
!
              do n=1,ndir
                 do nx=1,size(stepx)
                    xtrial(:) = par(:)%val + stepx(nx)*svdir(:,n)
!
                    call func(xtrial, ftrial,0)
                    nfcnev = nfcnev + 1
                    if (ftrial <= fopt) then
                        xopt(:) = xtrial
                        foptold = fopt
                        fopt = ftrial
                        newopt = .true.
                        write(0,*)'minimo trovato con step ',stepx(nx),n,fopt
                        if (print_graph) call fcnprint(ginfo,'B',xopt,fopt,temp,nfcnev,(m-1)*sc%ns+n)
                        cycle
                    endif
                 enddo
                 if (newopt) exit
              enddo
          !endif
          if (newopt) then
              call fcnprint(ginfo,ginfo%popt,xopt,fopt,temp,nfcnev,0)
              par(:)%val = xtrial(:)  
              fval = fopt
          else
!
!             Applica Metropolis
              prob = exprep(-(ftrial - fval)/temp)
              call random_number(random)
              if (random < prob) then
                  par(:)%val = xtrial(:)
                  fval = ftrial
              endif
              xgrad(:) = -xgrad(:)
          endif

      enddo loop_moves   ! end loop over Montecarlo moves at fixed temperature          
      !call timestep(1)
!
      call check_convergence(sc%eps,fval,fopt,fstar,quit)
      fval = fopt
      par(:)%val = xopt(:)
      call func(par%val, fval,0)      
      if (quit .or. kevent == 1 .or. kevent == 2) exit
      temp = sc%temp_fact*temp
!
   enddo loop_temp      ! end global do for changing temperature

!   CONTAINS
!
!   real function funcgrad(x)
!   real, dimension(:), intent(in) :: x
!   real :: fvalue
!   call func(x, fvalue,0)
!   funcgrad = fvalue 
!   end function funcgrad

!
   end subroutine sa_glob
!---------------------------------------------------------------------------------------


!---------------------------------------------------------------------------------------

   subroutine genera_direzioni(xgrad,sdir,svdir)
!
!  Genera altre direzioni discendenti dalla prima in svdir(:,1)
!
   real, dimension(:), intent(in)                          :: xgrad   ! gradiente
   real, dimension(size(xgrad)), intent(in)                :: sdir    ! prima direzione discendente
   real, dimension(:,:), intent(out) :: svdir   ! altre direzioni discendenti 
   real, dimension(size(xgrad))                            :: svet,slim
   integer                                                 :: i
   real                                                    :: sumvet
   integer                                                 :: npar
   real, dimension(size(xgrad))                            :: xrand
   real, dimension(size(xgrad))                            :: dx_lim,dx_exp
   real                                                    :: dtprod
   integer :: ndir
   integer :: ix,iy
   logical :: kpr = .false.
!
   svet(:) = xgrad(:) * sdir(:)
   sumvet = sum(svet)
   npar = size(xgrad)
   call random_number(xrand)
!
!  Stampe di controllo
   if (kpr) then
       dx_lim(:) = -(sumvet - svet(:))/xgrad(:)
       do i=1,npar
          slim(:) = sdir(:)
          write(6,*)'lim.',i,dot_product(slim,xgrad)
       enddo
       write(6,*)'xgrad          =',xgrad
       write(6,*)'prima dir. disc=',sdir
       write(6,*)'dx_exp         =',dx_exp
   endif
!
   dx_exp(:) = -xrand(:)*(sumvet - svet(:))/xgrad(:)
!
!  Genera una serie di vettori contenenti tutte le combinazioni di 0 e 1
   ndir = 0
   do i=1,npar
      ndir = ndir + 1
      svdir(:,ndir) = 0   !!!sdir(:)
      svdir(i,i) = 1.0   !!!slim(i)
   enddo
   do i=1,npar
      ix = i
      iy = i
      call expand_direction(svdir,ix,iy,ndir)
   enddo
!
!  Dove c'e' 1 metti sdir e dove c'e' zero metti dx_exp
   do i=1,ndir
      where(svdir(:,i) == 0) 
        svdir(:,i) = dx_exp(:)
      elsewhere
        svdir(:,i) = sdir(:)
      endwhere
      dtprod = dot_product(svdir(:,i),xgrad)
      if (dtprod > 0) then
          svdir(:,i) = - svdir(:,i)  ! iverti se non e' discendente
          if (kpr) write(6,*)'inversione per ',i
      endif
      if (kpr) then
          write(6,*)'dir. disc.',i,svdir(:,i)
          write(6,*)'dot_prod.',i,dot_product(svdir(:,i),xgrad)
      endif
   enddo
!
   end subroutine genera_direzioni

!---------------------------------------------------------------------------------------
   subroutine test_genera_dir()
   real, allocatable, dimension(:,:)                         :: sdisc
   integer :: ndir
   integer :: i
   integer :: npar
   integer :: ix,iy
!
   npar = 2
   allocate(sdisc(npar,2**npar - 1))
   ndir = 0
   do i=1,npar
      ndir = ndir + 1
      sdisc(:,ndir) = 0   !!!sdir(:)
      sdisc(i,i) = 1.0   !!!slim(i)
      write(0,*)'vet=',sdisc(:,ndir)
   enddo
   do i=1,npar
      ix = i
      iy = i
      call expand_direction(sdisc,ix,iy,ndir)
   enddo
!
   end subroutine test_genera_dir
!---------------------------------------------------------------------------------------

   recursive subroutine expand_direction(svet,vpos,xpos,ndir)
   real, dimension(:,:), intent(inout) :: svet ! npar x ndir
   integer,  intent(in)                :: vpos ! vettore di svet su cui applicare l'espansione
   integer,  intent(in)                :: xpos ! posizione da cui espandere
   integer, intent(inout)              :: ndir ! num. di direzioni calcolate
   integer                             :: i
   integer                             :: ix
!
   do i=xpos+1,size(svet,1)
      ndir = ndir + 1
      svet(:,ndir) = svet(:,vpos)
      svet(i,ndir) = svet(xpos,ndir)
      !write(0,*)vpos,xpos,'vet=',svet(:,ndir)
      !write(0,*)'vet=',svet(:,ndir)
      ix = ndir
      call expand_direction(svet,ix,i,ndir)
   enddo
!
   end subroutine expand_direction

!---------------------------------------------------------------------------------------

   subroutine sa_set_default(c)
!   
!  Inizializza le condizioni del SIMMAN algorithm
   type(SimAnn_Conditions_type), intent(inout) :: c
!   
!!!!!!   c%algo = 1                                 ! Algorithm type
   c%t_ini = 10.0                             ! Initial temperature
   c%temp_fact = 0.90                         ! Temperature reduction factor
   !c%eps = 1.0e-2                             ! Error tolerance for termination
   c%eps = 0.1                             ! Error tolerance for termination
   c%initconfig = 0                           ! Flag determining if the first configuration is random or read
   c%nt = 500                                 ! Number of iterations before temperature reduction
   c%num_temps = 100000                       ! Maximum number of temperatures in SA
   c%ns = 20                                  ! After ns*npar stp is updated
   c%neps = 4                                 ! Number of final function values used to decide upon termination
   c%autoti = .true.                          ! calcolo automatico di t_ini
   c%autont = .true.                          ! calcolo automatico di nt
   c%rand = -1
!   
   end subroutine sa_set_default
   
!---------------------------------------------------------------------------------------

   subroutine sa_set_temperature(sc,ginfo,par,func,funcs)
!
!  Calcola una temperatura inziale in modo tale che tutto lo spazio dei parametri venga esplorato
!  Il metodo e quello suggerito da Goffe nella descrizione del SIMMAN algorithm.
!  
   INTERFACE
      real function func(xpar,hpar)
      real, intent(in), dimension(:) :: xpar
      integer, intent(in)            :: hpar
      end function func

      subroutine funcs(xp,fval,grad,icod)
      real, dimension(:), intent(in)  :: xp
      real, intent(out)               :: fval
      real, dimension(:), intent(out), optional :: grad
      integer, intent(in)             :: icod
      end subroutine funcs
   END INTERFACE
   type(SimAnn_Conditions_type), intent(inout) :: sc     ! sc%t_ini in uscita
   type(opti_info_type), intent(in)            :: ginfo
   type(param_type), dimension(:), intent(in)  :: par
   type(SimAnn_Conditions_type)                :: sctmp
   type(param_type), dimension(size(par))      :: partmp
   real                                        :: tfin
   real, dimension(size(par))                  :: xopt
   real                                        :: fopt
   integer                                     :: nacc,nfcnev,nobds
   integer                                     :: npar
   integer                                     :: i
   character(len=200)                          :: string
   integer                                     :: kpr = 0
   integer                                     :: kevent
   type(opti_info_type)                        :: ginfotmp
!
!  Prima esegui una copia delle condizioni per SA
   sctmp = sc  
   ginfotmp = ginfo
!   
   npar = size(par)
!
!  Scegli un fattore di temperatura positivo. La temp. aumentera e con essa stp         
   sctmp%temp_fact = 1.5     
   sctmp%t_ini = 1.0         ! si sceglie una temperatura sufficientemente bassa
   sctmp%num_temps = 1
   sctmp%nt = 10
   ginfotmp%popt = ' '
!         
!  Richiama l'algoritmo piu volte (100 tentativi al massimo) finche stp diventa sufficientemente grande
   write(string,'(a)')'  Please wait ... automatic setting of starting temperature'
   if(kpr>0)write(6,'(/1x,a)')string
   do i=1,100      
      partmp = par       ! ripristina i parametri
      ginfotmp%nrunc = i
      call sa(partmp, sctmp, ginfotmp, tfin, xopt, fopt, nacc, nfcnev, nobds, kevent,func,funcs,-1.0)
!
!     stp deve essere 1 per tutti i parametri      
      write(string,'(a,i0,a,f0.4,a,f0.4)')'Attempt n.',i,' at temp=',sctmp%t_ini,' R=',sum(partmp%stp)/npar
      call PrintMess(trim(string))
      if(kpr>0)write(0,'(1x,a)')trim(string)
        !write(0,*)'stp=',partmp%stp
        !write(0,*)i,sctmp%t_ini,'diff=',abs(partmp%stp - (par%high - par%low))
      if (all(abs(partmp%stp - (par%high - par%low)) <= epsilon(1.0))) exit
      sctmp%t_ini = tfin*sctmp%temp_fact      
   enddo             
   sc%t_ini = sctmp%t_ini
!   
   end subroutine sa_set_temperature
   
!---------------------------------------------------------------------------------------

   subroutine sa_set_temperature_new(sc,gopti,par,func)
!
!  Determina la temperatura inziale calcolando la media dell'incremento dei
!  costi per un numero di transizioni casuali
!  
   USE SA_INTERFACE
   USE nr
   USE stat
   USE rand_mod
   USE messagemod
   USE General, only:QualePgm,SIRPGM
   INTERFACE
      subroutine func(xpar,h,hpar,gopti)
      USE progtype
      implicit none
      real, intent(in)    :: xpar(:)
      real, intent(out)   :: h
      integer, intent(in) :: hpar
      type(opti_info_type), intent(in) :: gopti
      end subroutine func
   END INTERFACE
   type(SimAnn_Conditions_type), intent(inout) :: sc     ! sc%t_ini in uscita
   type(opti_info_type), intent(in)            :: gopti
   type(param_type), dimension(:), intent(in)  :: par
   type(param_type), dimension(size(par))      :: partmp
   integer                                     :: nfcnev
   integer                                     :: npar
   integer                                     :: i,j,h
   character(len=200)                          :: string
   integer                                     :: nmoves
   real, dimension(:), allocatable             :: deltav,fvalv
   real            :: fval
   real, dimension(size(par))                  :: trialval
   real                                        :: random
   real                                        :: range
   real                                        :: ftrial
   real                                        :: ave,var
!
!  Prima esegui una copia delle condizioni per SA
   partmp(:) = par(:)
!   
   npar = size(par)
   nmoves = 1000 !sc%ns*npar*100
   nfcnev = 0
   call func(partmp%val, fval,0,gopti)
   allocate(deltav(nmoves*npar),fvalv(nmoves*npar))
   if (QualePgm == SIRPGM) then
       write(string,'(a)')'  Please wait ... automatic setting of starting temperature'
       call RunToolbar(1,string)
   else
       call clear_messages(ipos=2)
       call write_message('Please wait ... automatic setting of starting temperature',pos=2)
   endif
   call progress_bar_init(string)
   do j = 1, nmoves
      loop_par: do h=1,npar       ! loop on parameters
!      
         do i = 1, npar
            if (i == h) then
                call random_number(random)
                trialval(i) = partmp(i)%val + (random*2. - 1.) * partmp(i)%stp
            else
                trialval(i) = partmp(i)%val
            endif
!           
!           Check for parameters out of bounds
            range = partmp(i)%high - partmp(i)%low
            if (partmp(i)%bound == 0) then       ! Fixed boundaries
                if((trialval(i) < partmp(i)%low) .OR. (trialval(i) > partmp(i)%high)) then
                   call random_number(random)     
                   trialval(i) = partmp(i)%low + range*random
                endif
            else                              ! Periodic boundaries
                if(trialval(i) < partmp(i)%low) trialval(i) = trialval(i) + range
                if(trialval(i) > partmp(i)%high) trialval(i) = trialval(i) - range 
            endif
         enddo
!     
!        Evaluate the function with the trial point trialval and return as FP.
         call func(trialval, ftrial,h,gopti)
         partmp%val = trialval
         nfcnev = nfcnev + 1
         fvalv(nfcnev) = ftrial
      enddo loop_par  
      call progress_bar_update(cval=nfcnev,fval=nmoves*npar)
   enddo             
   call avevar(fvalv(:nfcnev),ave,var)
   sc%t_ini = sqrt(var)
   call progress_bar_close()
!   
   end subroutine sa_set_temperature_new
   
!---------------------------------------------------------------------------------------

   subroutine sa_print_conditions(sc,unt,title)
   type(SimAnn_Conditions_type), intent(in) :: sc
   integer, intent(in)                      :: unt
   character(len=*), intent(in)             :: title
!  
!!!!!!   write(unt,"(/8x,a)")'Simulated Annealing conditions'
   write(unt,"(/8x,a)")trim(title)
   write(unt,'(1x,a,f0.3)')'Initial temperature:          ',sc%t_ini
   write(unt,'(1x,a,f0.3)')'Temperature reduction factor: ',sc%temp_fact
   write(unt,'(1x,a,i0)')  'Number of cycles per temp:    ',sc%nt
   write(unt,'(1x,a,f0.4)')'Tolerance for termination:    ',sc%eps
!!!!   write(unt,'(1x,a)')     'Cost function:                '//trim(sc%cost_function_name)
!
   end subroutine sa_print_conditions
   
!---------------------------------------------------------------------------------------

   subroutine sa_print_intermidiate(t, fopt, nup, ndown, nrej, lnobds, nnew, unitp)
   real , intent(in)   :: fopt
   real, intent(in)    :: t
   integer, intent(in) :: nup, ndown, nrej, lnobds, nnew
   integer, intent(in) :: unitp
   integer             :: totmov
!
   totmov = nup + ndown + nrej
   write(unitp,'(/," Intermediate results before next temperature reduction",/)')
   write(unitp,'("  Current temperature:        ",f0.6)') t
   write(unitp,'("  Min function value:         ",f0.6)') fopt
   write(unitp,'("  Total moves:                ",I0)') totmov
   write(unitp,'("     downhill:                ",I0)') nup
   write(unitp,'("     accepted uphill:         ",I0)') ndown
   write(unitp,'("     rejected uphill:         ",I0)') nrej
   write(unitp,'("  Out of bounds trials:       ",I0)') lnobds
   write(unitp,'("  New maxima this temperature:",I0)') nnew
   write(unitp, '(" ")')
!   
   end subroutine sa_print_intermidiate
   
!---------------------------------------------------------------------------------------
  
   subroutine sa_print_final(fopt,nfcnev,nacc,nobds,tfin,unt)
   real, intent(in)    :: fopt
   integer, intent(in) :: nfcnev
   integer, intent(in) :: nacc
   integer, intent(in) :: nobds
   real, intent(in)    :: tfin
   integer, intent(in) :: unt
!
   write(unt,'(/1x,a,f0.6)')'Optimal function value:              ',fopt
   write(unt,'(1x,a,i0)')   'Number of function evaluations:      ',nfcnev
   write(unt,'(1x,a,i0)')   'Number of accepted evaluations:      ',nacc
   write(unt,'(1x,a,i0)')   'Number of out of bound evaluations:  ',nobds
   write(unt,'(1x,a,f0.5)') 'Final temperature:                   ',tfin
   if (err_san) then
       write(unt,'(1x,a)')'Error:'
       write(unt,'(1x,a)')trim(err_mess_san)
   endif
!
   end subroutine sa_print_final
   
!---------------------------------------------------------------------------------------

   subroutine sa_check(c)
!
!  Check all important variables of the SA algorithm
!   
   type(SimAnn_Conditions_type), intent(inout)   :: c          
!
!  Default values
   if (c%num_temps < 1) c%num_temps = 1
   if (c%t_ini < 0) c%t_ini=1   
!!!!!   if (len_trim(c%cost_function_name) == 0) c%cost_function_name = "Unnamed cost function"
!   
   end subroutine sa_check
   
!---------------------------------------------------------------------------------------

   function exprep(rval) result(exp_val)
!  This function replaces exp to avoid under- and overflows.
!  Note that the maximum and minimum values of
!  EXPREP are such that they has no effect on the algorithm.

   real ,intent(in) :: rval
   real             :: exp_val
   real             :: maxexp= huge(1.0)
!corr   real             :: maxarg=88.0  != log(huge(1.0))
!corr   real             :: minarg=-87.0 != log(tiny(1.0))
   real             :: maxarg=log(huge(1.0))
   real             :: minarg=log(tiny(1.0))
!
!   maxarg = log(huge(1.0))
!   minarg = log(tiny(1.0))
!   maxexp = huge(1.0)
!   write(6,*)maxarg,minarg,maxexp
   if (rval > maxarg) then
       exp_val = maxexp       
   elseif (rval < minarg ) then 
       exp_val = 0.0
   else
       exp_val = exp(rval)
   endif
!
   end function exprep
   
END MODULE simulated_anneal
