module spgext

   USE spginfom, only: spaceg_type
   implicit none

   type spgext_info_t
     integer           :: num         ! Space group number
     character(len=11) :: symb        ! Space group symbol
     character(len=20) :: exts        ! Extinction symbol
     real              :: fig         ! Figure of merit
     integer           :: nabs        ! Number of absent reflections not in overlap
     integer           :: nasym       ! Atoms in the asymmetric unit
     integer           :: freq_no     ! No. in CSD
     real              :: freq_perc   ! % in CSD
     integer           :: freq_rank   ! Rank in CSD
     logical           :: chiral      ! Chiral space group?
   end type spgext_info_t
!
!  Global variable for space group determination
   type(spgext_info_t), dimension(:), allocatable :: spglist
   type(spaceg_type)                              :: spglaue
   integer, dimension(:,:), allocatable           :: hklspg
   integer, dimension(:), allocatable             :: indspg
   real, dimension(:), allocatable                :: weispg

contains

   subroutine spg_from_exts(numsymb,fig,vnabs,spg,elem)
!
!  Generate spglist
!
   USE symm_table
   USE spginfom
   USE nr
   USE elements
   integer, dimension(:), intent(in)                           :: numsymb
   real, dimension(:), intent(in)                              :: fig
   integer, dimension(:), intent(in)                           :: vnabs
   type(spgext_info_t), dimension(:), allocatable, intent(out) :: spg
   type(element_type), dimension(:), allocatable, intent(in)   :: elem
   integer                                                     :: i,j
   integer                                                     :: inums,nsymb,ns,nspg
   type(spaceg_type)                                           :: sg
   real, dimension(:), allocatable                             :: vetord
   integer, dimension(:), allocatable                          :: iord
!
   nsymb = size(numsymb)
   if (nsymb == 0) return
   nspg = sum(extsy(numsymb)%nspg)
   allocate(spg(nspg)) 
   allocate(vetord(nspg),iord(nspg))
   ns = 0
   do i=1,nsymb
      inums = numsymb(i)
      do j=1,extsy(inums)%nspg
         ns = ns + 1
         spg(ns)%symb = sg_info(extsy(inums)%spgn(j))%hm2
         spg(ns)%fig = fig(i)
         spg(ns)%nabs = vnabs(i)
         spg(ns)%exts = extsy(inums)%symb
         spg(ns)%num = sg_info(extsy(inums)%spgn(j))%num
         sg%num = spg(ns)%num
         call sg%spg_frequency()
         spg(ns)%freq_no = sg%freq_no
         spg(ns)%freq_perc = sg%freq_perc
         spg(ns)%freq_rank = sg%freq_rank
         spg(ns)%chiral = sg%is_chiral()
         if (numelem(elem) > 0) then
             spg(ns)%nasym = nint(nasym_unit(elem,z_from_spg(sg%num),elem(1)%radtype==RX_SOURCE))
         else
             spg(ns)%nasym = 0
         endif
         vetord(ns) = i*1000+(1-spg(ns)%freq_perc)  ! 1-freq to obtain reverse order
      enddo
   enddo
!
!  Sort considering the frequency of space groups
   call indexx(vetord,iord)
   spg(:) = spg(iord)
!
   end subroutine spg_from_exts

!--------------------------------------------------------------------------------------

   integer function csys_from_exts(pos) result(csys)
!
!  Crystal system corresponding to an extinction symbol
!
   USE spginfom
   integer, intent(in) :: pos
   csys = csys_code_from_num(sg_info(extsy(pos)%spgn(1))%num)
   end function csys_from_exts

!--------------------------------------------------------------------------------------

   subroutine print_spglist(spg,kpr)
   type(spgext_info_t), dimension(:), allocatable, intent(in) :: spg
   integer, intent(in)                                        :: kpr
   integer                                                    :: ns
   character(len=3)                                           :: chiral
!
   if (size(spg) == 0) return
!
   write(kpr,'(2x,88("="),/2x,a/,2x,a,2x/,2x,88("="))')                    &
         '                            List of plausible space groups',  &
         'Space Group Extinction Symbol     FoM       Nabs   Nasym No.in CSD  % of CSD Rank Chiral'
!
   do ns=1,size(spg)
      if (spg(ns)%chiral) then
          chiral = 'yes'
      else
          chiral = 'no'
      endif
      write(kpr,'(2x,a,1x,a,f7.3,i8,i8,1x,i8,3x,f7.2,2x,i4,2x,a)')spg(ns)%symb,spg(ns)%exts,spg(ns)%fig,spg(ns)%nabs,   &
      spg(ns)%nasym,spg(ns)%freq_no,spg(ns)%freq_perc,spg(ns)%freq_rank,chiral
   enddo
   end subroutine print_spglist

!--------------------------------------------------------------------------------------

   subroutine save_spglist_bin(spg,unitbin)
   type(spgext_info_t), dimension(:), allocatable, intent(in) :: spg
   integer, intent(in)                                        :: unitbin
!
   if (allocated(spg)) then
       write(unitbin)size(spg),size(indspg)
       write(unitbin)spg,hklspg,indspg,weispg,spglaue
   else
       write(unitbin)0,0
   endif
!
   end subroutine save_spglist_bin

!--------------------------------------------------------------------------------------

   subroutine read_spglist_bin(spg,unitbin,err)
   USE errormod
   USE arrayutil
   type(spgext_info_t), dimension(:), allocatable, intent(out) :: spg
   integer, intent(in)                                         :: unitbin
   type(error_type), intent(out)                               :: err
   integer                                                     :: nspg,nref,ier
!
   read(unitbin,iostat=ier)nspg,nref
   if (ier /= 0) go to 10

   if (nspg > 0) then
       if (allocated(spg)) deallocate(spg)
       allocate(spg(nspg))
       call new_array(hklspg,[1,1],[nref,3])
       call new_array(indspg,nref)
       call new_array(weispg,nref)
       read(unitbin,iostat=ier)spg,hklspg,indspg,weispg,spglaue
       if (ier /= 0) go to 10
   endif
   return
!
10 call err%set("Error on reading space group")
!
   end subroutine read_spglist_bin

end module spgext
